

from epyk.core.js.packages import JsPackage
from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects


class ColumnApi(object):

  def __init__(self, report, varId):
    self.varId = varId
    self._report = report

  def sizeColumnsToFit(self, width):
    """
    Description:
    -----------
    Gets the grid to size the columns to the specified width in pixels, e.g. sizeColumnsToFix(900).
    To have the grid fit the columns to the grid's width, use the Grid API gridApi.sizeColumnsToFit() instead.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    Attributes:
    ----------
    :param columns:
    """
    width = JsUtils.jsConvertData(width, None)
    return JsObjects.JsVoid("%s.sizeColumnsToFit(%s)" % (self.varId, width))

  def getColumnGroup(self, name):
    """
    Description:
    -----------
    Returns the column group with the given name.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    Attributes:
    ----------
    :param name:
    """
    name = JsUtils.jsConvertData(name, None)
    return JsObjects.JsVoid("%s.getColumnGroup(%s)" % (self.varId, name))

  def getColumn(self, name):
    """
    Description:
    -----------
    Returns the column with the given colKey, which can either be the colId (a string) or the colDef (an object).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    Attributes:
    ----------
    :param name:
    """
    name = JsUtils.jsConvertData(name, None)
    return JsObjects.JsVoid("%s.getColumn(%s)" % (self.varId, name))

  def getColumnState(self):
    """
    Description:
    -----------
    Gets the state of the columns. Typically used when saving column state.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.getColumnState()" % self.varId)

  def setColumnState(self, columnState):
    """
    Description:
    -----------
    Sets the state of the columns from a previous state. Returns false if one or more columns could not be found.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    columnState = JsUtils.jsConvertData(columnState, None)
    return JsObjects.JsVoid("%s.setColumnState(%s)" % (self.varId, columnState))

  def resetColumnState(self):
    """
    Description:
    -----------
    Sets the state back to match the originally provided column definitions.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.resetColumnState()" % self.varId)

  def isPinning(self):
    """
    Description:
    -----------
    Returns true if pinning left or right, otherwise false.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.isPinning()" % self.varId)

  def isPinningLeft(self):
    """
    Description:
    -----------
    Returns true if pinning left, otherwise false.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.isPinningLeft()" % self.varId)

  def isPinningRight(self):
    """
    Description:
    -----------
    Returns true if pinning right, otherwise false.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.isPinningRight()" % self.varId)

  def setColumnVisible(self, colKey, visible):
    """
    Description:
    -----------
    Sets the visibility of a column. Key can be the column ID or Column object.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKey = JsUtils.jsConvertData(colKey, None)
    visible = JsUtils.jsConvertData(visible, None)
    return JsObjects.JsVoid("%s.setColumnVisible(%s. %s)" % (self.varId, colKey, visible))

  def setColumnsVisible(self, colKeys, visible):
    """
    Description:
    -----------
    Same as setColumnVisible, but provide a list of column keys.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKeys = JsUtils.jsConvertData(colKeys, None)
    visible = JsUtils.jsConvertData(visible, None)
    return JsObjects.JsVoid("%s.setColumnsVisible(%s. %s)" % (self.varId, colKeys, visible))

  def setColumnPinned(self, colKey, pinned):
    """
    Description:
    -----------
    Sets the column pinned / unpinned. Key can be the column ID, field, ColDef object or Column object.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKey = JsUtils.jsConvertData(colKey, None)
    pinned = JsUtils.jsConvertData(pinned, None)
    return JsObjects.JsVoid("%s.setColumnPinned(%s. %s)" % (self.varId, colKey, pinned))

  def setColumnsPinned(self, colKeys, pinned):
    """
    Description:
    -----------
    Sets the column pinned / unpinned. Key can be the column ID, field, ColDef object or Column object.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKeys = JsUtils.jsConvertData(colKeys, None)
    pinned = JsUtils.jsConvertData(pinned, None)
    return JsObjects.JsVoid("%s.setColumnsPinned(%s. %s)" % (self.varId, colKeys, pinned))

  def getColumnGroupState(self):
    """
    Description:
    -----------
    Gets the state of the column groups. Typically used when saving column group state.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.getColumnGroupState()" % self.varId)

  def autoSizeColumn(self, colKey):
    """
    Description:
    -----------
    Auto-sizes a column based on its contents.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKey = JsUtils.jsConvertData(colKey, None)
    return JsObjects.JsVoid("%s.autoSizeColumn(%s)" % (self.varId, colKey))

  def autoSizeColumns(self, colKeys):
    """
    Description:
    -----------
    Same as autoSizeColumn, but provide a list of column keys.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKeys = JsUtils.jsConvertData(colKeys, None)
    return JsObjects.JsVoid("%s.autoSizeColumns(%s)" % (self.varId, colKeys))

  def getDisplayNameForColumn(self, name):
    """
    Description:
    -----------
    Returns the display name for a column.
    Useful if you are doing your own header rendering and want the grid to work out if headerValueGetter is used, or if you are doing your own column management GUI, to know what to show as the column name.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    Attributes:
    ----------
    :param name:
    """
    name = JsUtils.jsConvertData(name, None)
    return JsObjects.JsVoid("%s.getDisplayNameForColumn(%s)" % (self.varId, name))

  def getAllColumns(self):
    """
    Description:
    -----------
    Returns all the columns, regardless of visible or not.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.getAllColumns()" % self.varId)

  def getAllGridColumns(self):
    """
    Description:
    -----------
    Returns all the grid columns, same as getAllColumns(), except a) it has the order of the columns that are presented in the grid and b) it's after the 'pivot' step, so if pivoting, has the value columns for the pivot.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.getAllGridColumns()" % self.varId)

  def getPrimaryColumns(self):
    """
    Description:
    -----------
    Returns the grid's primary columns.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.getPrimaryColumns()" % self.varId)

  def getSecondaryColumns(self):
    """
    Description:
    -----------
    Returns the grid's secondary columns.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.getSecondaryColumns()" % self.varId)

  def getAllDisplayedVirtualColumns(self):
    """
    Description:
    -----------
    Same as getAllGridColumns(), except only returns rendered columns, i.e. columns that are not within the viewport and therefore not rendered, due to column virtualisation, are not displayed.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.getAllDisplayedVirtualColumns()" % self.varId)

  def moveColumn(self, colKey, toIndex):
    """
    Description:
    -----------
    Moves a column to toIndex. The column is first removed, then added at the toIndex location, thus index locations will change to the right of the column after the removal.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKey = JsUtils.jsConvertData(colKey, None)
    toIndex = JsUtils.jsConvertData(toIndex, None)
    return JsObjects.JsVoid("%s.moveColumn(%s, %s)" % (self.varId, colKey, toIndex))

  def moveColumns(self, colKeys, toIndex):
    """
    Description:
    -----------
    Moves a column to toIndex. The column is first removed, then added at the toIndex location, thus index locations will change to the right of the column after the removal.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKeys = JsUtils.jsConvertData(colKeys, None)
    toIndex = JsUtils.jsConvertData(toIndex, None)
    return JsObjects.JsVoid("%s.moveColumns(%s, %s)" % (self.varId, colKeys, toIndex))

  def setColumnAggFunc(self, column, aggFunc):
    """
    Description:
    -----------
    Sets the agg function for a column. aggFunc can be one of 'min' | 'max' | 'sum'.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    column = JsUtils.jsConvertData(column, None)
    aggFunc = JsUtils.jsConvertData(aggFunc, None)
    return JsObjects.JsVoid("%s.setColumnAggFunc(%s, %s)" % (self.varId, column, aggFunc))

  def setColumnWidth(self, colKey, newWidth, finished=True):
    """
    Description:
    -----------
    Sets the column width on a single column. The finished flag gets included in the resulting event and not used internally by the grid.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    colKey = JsUtils.jsConvertData(colKey, None)
    newWidth = JsUtils.jsConvertData(newWidth, None)
    finished = JsUtils.jsConvertData(finished, None)
    return JsObjects.JsVoid("%s.setColumnWidth(%s, %s, %s)" % (self.varId, colKey, newWidth, finished))

  def setColumnWidths(self, columnWidths, finished=True):
    """
    Description:
    -----------
    Sets the column width on a single column. The finished flag gets included in the resulting event and not used internally by the grid.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    columnWidths = JsUtils.jsConvertData(columnWidths, None)
    finished = JsUtils.jsConvertData(finished, None)
    return JsObjects.JsVoid("%s.setColumnWidth(%s, %s)" % (self.varId, columnWidths, finished))

  def custom(self, func_nam, *argv):
    """
    Description:
    ------------
    Generic function to call any missing function form a package.
    This will automatically convert the object to JavaScript and also put the right object reference.

    Attributes:
    ----------
    :param func_nam: String. The function name
    :param argv: Objects. Optional. The function arguments on the JavasScript side
    """
    js_args = []
    for arg in argv:
      js_args.append(str(JsUtils.jsConvertData(arg, None)))
    return JsObjects.JsObject.JsObject.get("%s.%s(%s)" % (self.varId, func_nam, ", ".join(js_args)))


class AgGrid(JsPackage):
  lib_alias = {"js": "ag-grid", "css": "ag-grid"}

  def sizeColumnsToFit(self):
    """
    Description:
    -----------
    Gets columns to adjust in size to fit the grid horizontally

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/
    """
    return JsObjects.JsVoid("%s.api.sizeColumnsToFit()" % self.varId)

  @property
  def columnApi(self):
    """

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-definitions/
    """
    return ColumnApi(self.src, "%s.columnApi" % self.varId)

  def setColumnDefs(self, colDefs):
    """
    Description:
    -----------
    Call to set new column definitions. The grid will redraw all the column headers, and then redraw all of the rows.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.setColumnDefs(%s)" % (self.varId, JsUtils.jsConvertData(colDefs, None)))

  def setRowData(self, rows):
    """
    Description:
    -----------
    Set rows.

    Related Pages:

			https://www.ag-grid.com/javascript-grid-api/

		Attributes:
    ----------
    :param rows:
    """
    return JsObjects.JsVoid("%s.api.setRowData(%s)" % (self.varId, JsUtils.jsConvertData(rows, None)))

  def applyTransaction(self, transaction):
    """
    Description:
    -----------
    Update row data. Pass a transaction object with lists for add, remove and update.

    Related Pages:

			https://www.ag-grid.com/javascript-grid-api/

    Attributes:
    ----------
    :param transaction:
    """
    return JsObjects.JsVoid("%s.api.applyTransaction(%s)" % (self.varId, JsUtils.jsConvertData(transaction, None)))

  def applyTransactionAsync(self, transaction, callback):
    """
    Description:
    -----------
    Same as applyTransaction except executes asynchronous for efficiency.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/

    Attributes:
    ----------
    :param transaction:
    :param callback:
    """
    return JsObjects.JsVoid("%s.api.applyTransaction(%s, %s)" % (self.varId, JsUtils.jsConvertData(transaction, None), callback))

  def getDisplayedRowCount(self):
    """
    Description:
    -----------
    Returns the total number of displayed rows.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getDisplayedRowCount()" % self.varId)

  def getFirstDisplayedRow(self):
    """
    Description:
    -----------
    Get the index of the first displayed row due to scrolling (includes invisible rendered rows in the buffer).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getFirstDisplayedRow()" % self.varId)

  def getLastDisplayedRow(self):
    """
    Description:
    -----------
    Get the index of the last displayed row due to scrolling (includes invisible rendered rows in the buffer).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getLastDisplayedRow()" % self.varId)

  def hideColumns(self, columns):
    """

    :param columns:
    """
    return JsObjects.JsVoid("%(varId)s.columnApi.setColumnsVisible(%(cols)s, false)" % {'varId': self.varId, 'cols': JsUtils.jsConvertData(columns, None)})

  def showColumns(self, columns):
    """

    :param columns:
    """
    return JsObjects.JsVoid("%(varId)s.columnApi.setColumnsVisible(%(cols)s, true)" % {'varId': self.varId, 'cols': JsUtils.jsConvertData(columns, None)})

  def hideColumn(self, column):
    """

    https://www.ag-grid.com/javascript-grid-column-definitions/

    :param columns:
    """
    return JsObjects.JsVoid("%(varId)s.columnApi.setColumnVisible(%(cols)s, false)" % {'varId': self.varId, 'cols': JsUtils.jsConvertData(column, None)})

  def showColumn(self, column):
    """

    https://www.ag-grid.com/javascript-grid-column-definitions/

    :param columns:
    """
    return JsObjects.JsVoid("%(varId)s.columnApi.setColumnVisible(%(cols)s, true)" % {'varId': self.varId, 'cols': JsUtils.jsConvertData(column, None)})

  def getRowNode(self, row_id):
    """
    Description:
    -----------
    Returns the row node with the given ID. The row node ID is the one you provide with the callback getRowNodeId(data), otherwise the ID is a number auto generated by the grid when the row data is set.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    row_id = JsUtils.jsConvertData(row_id, None)
    return JsObjects.JsVoid("%s.api.getRowNode(%s)" % (self.varId, row_id))

  def getDisplayedRowAtIndex(self, index):
    """
    Description:
    -----------
    Returns the displayed rowNode at the given index.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    index = JsUtils.jsConvertData(index, None)
    return JsObjects.JsVoid("%s.api.getRowNode(%s)" % (self.varId, index))

  def getDisplayedRowCount(self):
    """
    Description:
    -----------
    Returns the total number of displayed rows.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getDisplayedRowCount()" % self.varId)

  def getFirstDisplayedRow(self):
    """
    Description:
    -----------
    Get the index of the first displayed row due to scrolling (includes invisible rendered rows in the buffer).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getFirstDisplayedRow()" % self.varId)

  def getLastDisplayedRow(self):
    """
    Description:
    -----------
    Get the index of the last displayed row due to scrolling (includes invisible rendered rows in the buffer).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getLastDisplayedRow()" % self.varId)

  def selectAll(self):
    """
    Description:
    -----------
    Select all rows (even rows that are not visible due to grouping being enabled and their groups not expanded).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.selectAll()" % self.varId)

  def deselectAll(self):
    """
    Description:
    -----------
    Clear all row selections.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.deselectAll()" % self.varId)

  def selectAllFiltered(self):
    """
    Description:
    -----------
    Select all filtered rows.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.selectAllFiltered()" % self.varId)

  def deselectAllFiltered(self):
    """
    Description:
    -----------
    Clear all filtered selections.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.deselectAllFiltered()" % self.varId)

  def getSelectedNodes(self):
    """
    Description:
    -----------
    Returns a list of selected nodes. Getting the underlying node (rather than the data) is useful when working with tree / aggregated data, as the node can be traversed.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getSelectedNodes()" % self.varId)

  def getSelectedRows(self):
    """
    Description:
    -----------
    Returns a list of selected rows (i.e. row data that you provided).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getSelectedRows()" % self.varId)

  def getCellRanges(self):
    """
    Description:
    -----------
    Returns the list of selected cell ranges.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.getCellRanges()" % self.varId)

  def clearRangeSelection(self):
    """
    Description:
    -----------
    Clears the selected range.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.clearRangeSelection()" % self.varId)

  def refreshCells(self, params):
    """
    Description:
    -----------
    Performs change detection on all cells, refreshing cells where required.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    params = JsUtils.jsConvertData(params, None)
    return JsObjects.JsVoid("%s.api.refreshCells(%s)" % (self.varId, params))

  def redrawRows(self, params):
    """
    Description:
    -----------
    Remove a row from the DOM and recreate it again from scratch.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    params = JsUtils.jsConvertData(params, None)
    return JsObjects.JsVoid("%s.api.redrawRows(%s)" % (self.varId, params))

  def refreshHeader(self):
    """
    Description:
    -----------
    Redraws the header. Useful if a column name changes, or something else that changes how the column header is displayed.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.refreshHeader(%s)" % (self.varId))

  def flashCells(self, params):
    """
    Description:
    -----------
    Flash rows, columns or individual cells. See Flashing Cells.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    params = JsUtils.jsConvertData(params, None)
    return JsObjects.JsVoid("%s.api.flashCells(%s)" % (self.varId, params))

  def clearFocusedCell(self):
    """
    Description:
    -----------
    Clears the focused cell.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.clearFocusedCell()" % (self.varId))

  def tabToNextCell(self):
    """
    Description:
    -----------
    Navigates the grid focus to the next cell, as if tabbing.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.tabToNextCell()" % (self.varId))

  def tabToPreviousCell(self):
    """
    Description:
    -----------
    Navigates the grid focus to the previous cell, as if shift-tabbing.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.tabToPreviousCell()" % (self.varId))

  def showLoadingOverlay(self):
    """
    Description:
    -----------
    Show the 'loading' overlay.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.showLoadingOverlay()" % (self.varId))

  def showNoRowsOverlay(self):
    """
    Description:
    -----------
    Show the 'no rows' overlay.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.showNoRowsOverlay()" % (self.varId))

  def hideOverlay(self):
    """
    Description:
    -----------
    Hides the overlay if showing.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.hideOverlay()" % (self.varId))

  def destroy(self):
    """
    Description:
    -----------
    Will destroy the grid and release resources.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.destroy()" % (self.varId))

  def resetRowHeights(self):
    """
    Description:
    -----------
    Tells the grid to recalculate the row heights.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.resetRowHeights()" % (self.varId))

  def paginationIsLastPageFound(self):
    """
    Description:
    -----------
    Returns true when the last page is known; this will always be the case if you are using the Client-Side Row Model for pagination.
    Returns false when the last page is not known; this only happens when using Infinite Scrolling Row Model.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationIsLastPageFound()" % (self.varId))

  def copySelectedRangeToClipboard(self, includeHeaders):
    """
    Description:
    -----------
    Copies the selected ranges to the clipboard.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    includeHeaders = JsUtils.jsConvertData(includeHeaders, None)
    return JsObjects.JsVoid("%s.api.copySelectedRangeToClipboard(%s)" % (self.varId, includeHeaders))

  def copySelectedRangeDown(self):
    """
    Description:
    -----------
    Copies the selected range down, similar to Ctrl + D in Excel.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.copySelectedRangeDown()" % self.varId)

  def paginationGetPageSize(self):
    """
    Description:
    -----------
    Returns how many rows are being shown per page.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGetPageSize()" % (self.varId))

  def paginationSetPageSize(self, newPageSize):
    """
    Description:
    -----------
    Sets the paginationPageSize to newPageSize, then re-paginates the grid so the changes are applied immediately.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    newPageSize = JsUtils.jsConvertData(newPageSize, None)
    return JsObjects.JsVoid("%s.api.paginationSetPageSize(%s)" % (self.varId, newPageSize))

  def paginationGetCurrentPage(self):
    """
    Description:
    -----------
    Returns the 0-based index of the page which is showing.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGetCurrentPage()" % (self.varId))

  def paginationGetTotalPages(self):
    """
    Description:
    -----------
    Returns the total number of pages. Returns null if paginationIsLastPageFound() == false.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGetTotalPages()" % (self.varId))

  def paginationGetRowCount(self):
    """
    Description:
    -----------
    The total number of rows. Returns null if paginationIsLastPageFound() == false.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGetRowCount()" % (self.varId))

  def paginationGoToPage(self, pageNumber):
    """
    Description:
    -----------
    Goes to the specified page. If the page requested doesn't exist, it will go to the last page.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    pageNumber = JsUtils.jsConvertData(pageNumber, None)
    return JsObjects.JsVoid("%s.api.paginationGoToPage(%s)" % (self.varId, pageNumber))

  def paginationGoToNextPage(self):
    """
    Description:
    -----------
    Shorthands for goToPage(relevantPageNumber).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGoToNextPage()" % self.varId)

  def paginationGoToPreviousPage(self):
    """
    Description:
    -----------
    Shorthands for goToPage(relevantPageNumber).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGoToPreviousPage()" % self.varId)

  def paginationGoToFirstPage(self):
    """
    Description:
    -----------
    Shorthands for goToPage(relevantPageNumber).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGoToFirstPage()" % self.varId)

  def paginationGoToLastPage(self):
    """
    Description:
    -----------
    Shorthands for goToPage(relevantPageNumber).

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    return JsObjects.JsVoid("%s.api.paginationGoToLastPage()" % self.varId)

  def setSideBarVisible(self, show):
    """
    Description:
    -----------
    Tells the grid to recalculate the row heights.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-api/
    """
    show = JsUtils.jsConvertData(show, None)
    return JsObjects.JsVoid("%s.api.setSideBarVisible(%s)" % (self.varId, show))

  def getSortModel(self):
    """
    Description:
    -----------
    Returns the sort state.

    Related Pages:

      https://www.ag-grid.com/javascript-grid-column-api/

    """
    return JsObjects.JsVoid("%s.api.getSortModel()" % self.varId)




from epyk.core.js import JsUtils


class Effects(object):

  def __init__(self, report, htmlObj, ovrs_attrs=None):
    self._report, self._htmlObj = report, htmlObj
    if ovrs_attrs is not None:
      self.attrs = dict(self.attrs)
      self.attrs.update(ovrs_attrs)

  def get_attrs(self):
    """
    Description:
    ------------
    Return the effect CSS attributes
    """
    return self.attrs

  def glow(self, color, radius=50):
    """
    Description:
    ------------
    Use the text-shadow property to create the neon light effect, and then use animation together with keyframes to add the repeatedly glowing effect

    Related Pages:

			https://www.w3schools.com/howto/howto_css_glowing_text.asp

    Attributes:
    ----------
    :param color: String. The color to use fin the effect
    :param radius: Integer. The lenght of the radius to display in the animate
    """
    keyframe_name = "glow_%s" % color
    self._htmlObj.style.css.animation = "%s 1s ease-in-out infinite alternate" % keyframe_name
    color_effects = []
    for i in range(1, int(radius / 10)+1):
      color_effects.append("0 0 %s0px %s" % (i, color))
    if not color_effects:
      color_effects.append("0 0 %spx %s" % (radius, color))
    attrs = {"from": {"text-shadow": color_effects}, "to": {"text-shadow": color_effects}}
    self._htmlObj.style.css_class.keyframes(keyframe_name, attrs)
    return self

  def blink(self, duration=1):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param duration:
    """
    keyframe_name = "blink_%s" % duration
    self._htmlObj.style.css.animation = "%s %ss ease-in-out infinite alternate" % (keyframe_name, duration)
    attrs = {"from": {"opacity": 0}, "to": {"opacity": 1}}
    self._htmlObj.style.css_class.keyframes(keyframe_name, attrs)
    return self

  def shiny_text(self, color):
    """
    Description:
    ------------
    Use the text-shadow property to create the neon light effect, and then use animation together with keyframes to add the repeatedly glowing effect

    Related Pages:

			https://www.w3schools.com/howto/howto_css_glowing_text.asp

    Attributes:
    ----------
    :param color: String. The color to use fin the effect
    """
    keyframe_name = "shiny_text"
    self._htmlObj.style.css.animation = "%s 1s ease-in-out infinite alternate" % keyframe_name
    attrs = {"from": {"color": color}, "to": {"color": "none"}}
    self._htmlObj.style.css_class.keyframes(keyframe_name, attrs)
    return self

  def shiny_border(self, color, duration=1):
    """
    Description:
    ------------
    Use the text-shadow property to create the neon light effect, and then use animation together with keyframes to add the repeatedly glowing effect

    Related Pages:

			https://www.w3schools.com/howto/howto_css_glowing_text.asp

    Attributes:
    ----------
    :param color: String. The color to use fin the effect
    :param duration:
    """
    keyframe_name = "shiny_border"
    self._htmlObj.style.css.animation = "%s %ss ease-in-out infinite alternate" % (keyframe_name, duration)
    attrs = {"from": {"border": "1px solid %s" % color}, "to": {"border": "1px solid %s" % color}}
    self._htmlObj.style.css_class.keyframes(keyframe_name, attrs)
    return self

  def spin(self, duration):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param duration:
    """
    keyframe_name = "spin_%s" % duration
    self._htmlObj.style.css.animation = "%s %ss ease-in-out infinite alternate" % (keyframe_name, duration)
    attrs = {"from": {"transform": "rotate(0deg)"}, "to": {"transform": "rotate(360deg)"}}
    self._htmlObj.style.css_class.keyframes(keyframe_name, attrs)
    return self

  def translate(self, duration):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param duration:

    """
    keyframe_name = "trans_%s" % duration
    self._htmlObj.style.css.animation = "%s %ss ease-in-out infinite alternate" % (keyframe_name, duration)
    attrs = {"from": {"transform": "translateY(-100%)"}, "to": {"transform": "translateY(0)"}}
    self._htmlObj.style.css_class.keyframes(keyframe_name, attrs)
    return self

  def animate(self, name, targ_css_attrs, orig_css_attrs=None, delay=0, duration=1, timing_function="ease-in-out",
              iteration_count="infinite", directions="alternate"):
    """
    Description:
    ------------
    Use the text-shadow property to create the neon light effect, and then use animation together with keyframes to add the repeatedly glowing effect

    Attributes:
    ----------
    htmlObj.style.effects.animate("pink", {"color": "blue", "width": "400px"})

    Related Pages:

			https://www.w3schools.com/cssref/css_animatable.asp

    :param name: String. The animation name
    :param targ_css_attrs: Dictionary. The different CSS attributes to animate
    :param orig_css_attrs: Dictionary. The initial state of the attributes to animate
    :param delay: Integer. The delay in second before starting the animation
    :param timing_function: String. The animation-timing-function property specifies the speed curve of the animation.
    :param duration: Integer. The animation duration in second
    :param iteration_count: Integer. The animation count
    :param directions: String. The animation-direction property specifies whether an animation should be played forwards, backwards or in alternate cycles.
    """
    name = JsUtils.getJsValid(name, fail=False)
    keyframe_name = "animate_%s" % name
    self._htmlObj.style.css.animation = "%s %ss %s %ss %s %s" % (keyframe_name, duration, timing_function, delay, iteration_count, directions)
    if orig_css_attrs is None:
      attrs = {"to": targ_css_attrs}
    else:
      attrs = {"from": orig_css_attrs, "to": targ_css_attrs}
    self._htmlObj.style.css_class.keyframes(keyframe_name, attrs)
    return self




from epyk.core import html


class Code(object):

  def __init__(self, context):
    self.context = context

  def css(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    CSS Text editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'css', 'matchBrackets': True, 'styleActiveLine': True, 'autoRefresh': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def xml(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    XML Text editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'xml', 'matchBrackets': True, 'styleActiveLine': True, 'autoRefresh': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def sql(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    SQL Text editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'sql', 'matchBrackets': True, 'styleActiveLine': True, 'autoRefresh': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def r(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    R Text editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'r', 'matchBrackets': True, 'styleActiveLine': True, 'autoRefresh': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def python(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    Python Text editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'python', 'matchBrackets': True, 'styleActiveLine': True, 'autoRefresh': False}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def javascript(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    Javascript Text editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'javascript', 'autoRefresh': True, 'styleActiveLine': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def markdown(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    Markdown Text editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'markdown', 'autoRefresh': True, 'styleActiveLine': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def rst(self, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    RestructuredText editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': 'rst', 'autoRefresh': True, 'styleActiveLine': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code

  def code(self, language, text="", color=None, width=(90, '%'), height=(200, 'px'), htmlCode=None, options=None, helper=None, profile=None):
    """
    Description:
    ------------
    Generic code editor

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTextEditor.Code`

    Related Pages:

			https://codemirror.net/index.html

    Attributes:
    ----------
    :param language: String. The language
    :param text: String. The text
    :param color: String. The color code
    :param width: Tuple. The with details in the format(value, unit)
    :param height: Tuple. The height details in the format(value, unit)
    :param htmlCode: String. The unique component ID
    :param options: Dictionary. The object properties
    :param helper: String. Optional. The helper message
    :param profile: Dictionary. Profiling details
    """
    dflt_options = {"lineNumbers": True, 'mode': language, 'autoRefresh': True, 'styleActiveLine': True}
    if options is not None:
      dflt_options.update(options)
    html_code = html.HtmlTextEditor.Code(self.context.rptObj, text, color, width, height, htmlCode, dflt_options, helper, profile)
    return html_code




import json

from epyk.core.html import Html

# The list of CSS classes
# from epyk.core.css.styles import GrpCls
# from epyk.core.css.styles import CssGrpClsText


class ExternalLink(Html.Html):
  name = 'External link'

  def __init__(self, report, text, url, icon, helper, height, decoration, options, profile):
    super(ExternalLink, self).__init__(report, {"text": text, "url": url}, css_attrs={'height': height}, profile=profile)
    # Add the internal components icon and helper
    self.add_icon(icon)
    self.add_helper(helper)
    if not 'url' in self.val:
      self.val['url'] = self.val['text']
    if 'target' in options:
      self.set_attrs(name="target", value=options['target'])
    self.decoration, self.__url = decoration, {}

  @property
  def _js__builder__(self):
    return 'htmlObj.innerHTML = data.text; htmlObj.href = data.url'

  def no_decoration(self, color=None):
    """
    Property to remove the list default style
    """
    self.style.css.text_decoration = None
    self.style.list_style_type = None
    if color is None:
      color = self._report.theme.greys[-1]
    self.style.css.color = color
    return self

  def build(self, data=None, options=None, profile=False):
    if not isinstance(data, dict):
      data = {"text": data}
    if "url" not in data:
      data["url"] = self.val["url"]
    return super(ExternalLink, self).build(data, options, profile)

  def __str__(self):
    self.set_attrs(name="href", value=self.val['url'])
    return '<a %s>%s</a>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val['text'], self.helper)


class DataLink(Html.Html):
  name = 'Data link'
  # _grpCls = CssGrpClsText.CssClassHref

  def __init__(self, report, text, value, width, height, format, profile):
    super(DataLink, self).__init__(report, {"text": text, 'value': value}, profile=profile,
                                   css_attrs={"width": width, 'height': height})
    self.format = format

  @property
  def no_decoration(self):
    """
    Property to remove the list default style
    """
    self.style.css.text_decoration = None
    self.style.css.list_style_type = None
    return self

  @property
  def _js__builder__(self):
    return '''
      var b = new Blob([data.value]); htmlObj.href = URL.createObjectURL(b);
      htmlObj.innerHTML = data.text'''

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<a %(attr)s href="#" download="Download.%(format)s" type="text/%(format)s">%(val)s</a>' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'val': self.val['text'], 'format': self.format}




from epyk.core.html import Html
from epyk.core.html import Defaults

from epyk.core.js import JsUtils
from epyk.core.js.packages import JsQuery

from epyk.core.html.options import OptTrees

from epyk.core.css.styles import GrpClsList


class Tree(Html.Html):
  name = 'List Expandable'
  requirements = ('font-awesome', )

  def __init__(self, report, records, color, width, height, htmlCode, helper, option, profile):
    super(Tree, self).__init__(report, records, css_attrs={"width": width, 'height': height})
    option['is_root'] = True
    option['icon_open'] = "fas fa-folder-open"
    option['style'] = {"list-style": 'none', 'margin-left': '8px', 'padding-left': 0}
    self.__options = OptTrees.OptionsTree(self, option)
    self.css(option['style'])
    self._jsStyles['click_node'] = None

  @property
  def options(self):
    """
    Description:
    -----------

    :rtype: OptTrees.OptionsTree
    """
    return self.__options

  @property
  def _js__builder__(self):
    return ''' if(options){htmlObj.innerHTML = ''; options.is_root = false};
      data.forEach(function(item, i){
        var li = document.createElement("li");
        var a = document.createElement("a");
        if(typeof item.items !== 'undefined'){
          var ul = document.createElement("ul"); 
          for(const attr in options.style){ul.style[attr] = options.style[attr]};
          %(builder)s(ul, item.items, options);
          var icon = document.createElement("i"); icon.style.marginRight = '5px';
          icon.onclick = function(){ 
            var ulDisplay = this.parentNode.querySelector('ul').style.display;
            if(ulDisplay == 'none'){ this.parentNode.querySelector('ul').style.display = 'block'}
            else{this.parentNode.querySelector('ul').style.display = 'none'}
          };

          options.icon_open.split(" ").forEach(function(s){icon.classList.add(s)});
          var span = document.createElement("span"); 
          span.innerHTML = item.value;
          a.appendChild(icon); a.appendChild(span); a.appendChild(ul);
        } else {
          a.innerHTML = item.value;
          a.style.paddingLeft = '18px';
        }
       
        li.appendChild(a);
        htmlObj.appendChild(li)
      })''' % {"builder": self.builder_name}

  def clickNode(self, jsFncs, profile=False):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    if not isinstance(jsFncs, list):
      jsFncs = []

    self._jsStyles['click_node'] = "function(event, value){%s} " % JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return self

  def click(self, jsFncs, profile=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    if not isinstance(jsFncs, list):
      jsFncs = []

    self._jsStyles['click_leaf'] = "function(event, value){%s} " % JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return self

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<ul %s></ul>' % self.get_attrs(pyClassNames=self.style.get_classes())


class TreeInput(Tree):

  def set(self, ul, data):
    """

    :param ul:
    :param data:
    """
    for l in data:
      if l.get('items') is not None:
        sub_l = self._report.ui.list()
        sub_l.options.managed = False
        ul.add_item(sub_l)[-1].no_decoration
        ul[-1].add_label(l['label'], css={"color": l.get('color', 'none')})
        ul[-1].add_icon(self.options.icon_open if self.options.expanded else self.options.icon_close)
        if not self.options.expanded:
          sub_l.css({"display": 'none'})
        ul[-1].icon.click([
          ul[-1].val.dom.toggle(),
          ul[-1].icon.dom.switchClass(self.options.icon_close.split(" ")[-1], self.options.icon_open.split(" ")[-1])])
        self.set(sub_l, l.get('items'))
      else:
        ul.add_item(self._report.ui.text(l['label']).editable().css({"width": 'none', 'min-width': 'none'}))[-1].no_decoration
        ul[-1].css({"color": l.get('color', 'none')})
    return self


class DropDown(Html.Html):
  requirements = ('bootstrap', 'jquery')
  name = 'DropDown Select'

  def __init__(self, report, data, text, width, height, htmlCode, helper, options, profile):
    self.__options = {}
    super(DropDown, self).__init__(report, text, htmlCode=htmlCode, profile=profile, css_attrs={"width": width, "height": height})
    self._vals, self.text = data, text
    self.css({'padding': 0, 'margin': "1px", "display": "block", "z-index": 10, 'cursor': 'pointer', 'position': 'relative'})
    self._jsStyles = {"a": {'text-decoration': 'none', 'line-height': '%spx' % Defaults.LINE_HEIGHT, 'padding': '0 10px', "width": '%spx' % options.get("width")},
                      "ul": {"left": "%spx" % options.get("width")}}
    self.__options = OptTrees.OptDropDown(self, options)

  @property
  def style(self):
    """
    Description:
    -----------

    :rtype: GrpClsList.ClassDropDown
    """
    if self._styleObj is None:
      self._styleObj = GrpClsList.ClassDropDown(self)
    return self._styleObj

  @property
  def options(self):
    """
    Description:
    -----------
    Property to set all the possible object for a dropdown

    :rtype: OptTrees.OptDropDown
    """
    return self.__options

  def click(self, jsFncs, profile=False, source_event=None):
    if not isinstance(jsFncs, list):
      jsFncs = []
    self._jsStyles['click'] = "function(event, value){%s} " % JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return self

  @property
  def _js__builder__(self):
    return ''' 
      var jqHtmlObj = %(jqId)s; if(options.clearDropDown){jqHtmlObj.empty()};
      data.forEach(function(rec){
        if (rec.items != undefined) {
          var li = $('<li class="dropdown" style="list-style-type:none;display:list-item;text-align:-webkit-match-parent"></li>'); var a = $('<a tabindex=-1>'+ rec.value +'<span class="caret"></span></a>');
          li.append(a); var ul = $('<ul class="submenu"></ul>'); ul.css(options.ul); options.clearDropDown = false; a.css(options.a);
          %(pyCls)s(ul, rec.items, options); li.append(ul); jqHtmlObj.append(li);
        } else {
          if (rec.url == undefined){var a = $('<a href="#">'+ rec.value +'</a>')}
          else {var a = $('<a href="'+ rec.url +'">'+ rec.value +'</a>')}; a.css(options.a);
          a.click(function(event){const value = a.html(); options.click(event, value)} );
          var li = $('<li style="list-style-type:none;display:list-item;text-align:-webkit-match-parent"></li>'); li.append(a); jqHtmlObj.append(li)}
      })''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False), "pyCls": self.__class__.__name__}

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return "<ul %s></ul>" % (self.get_attrs(pyClassNames=self.style.get_classes()))




from epyk.core.js.html import JsHtml


class JsHtmlD3(JsHtml.JsHtml):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    super(JsHtmlD3, self).__init__(htmlObj, varName, setVar, isPyData, report)
    self.varName = "d3.select('#%s')" % self.htmlCode





from epyk.core import html


class Steppers(object):

  def __init__(self, context):
    self.context = context

  def stepper(self, records, width=(100, '%'), height=(70, 'px'), color=None, options=None):
    """
    Description:

    Entry point for the stepper object

    Usage::

      rptObj.ui.stepper([
        {"value": 'test 1', "status": 'success', 'label': 'test'},
        {"value": 'test 2', "status": 'error'},
        {"value": 'test 3', "status": 'pending'}])

    Attributes:
    ----------
    :param records: A list with the different steps defined in the workflow
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param color: Optional.
    :param options: Optional. Specific Python options available for this component
    """
    dft_options = {"line": True}
    if options is not None:
      dft_options.update(options)
    st = html.HtmlStepper.Stepper(self.context.rptObj, records, width, height, color, dft_options)
    st.style.add_classes.div.stepper()
    st.style.css_class.media({'.cssdivstepper li': {"float": None, 'width': '100%'},
                               '.cssdivstepper li line': {'stroke-width': 0},
                               '.cssdivstepper li [name=label]': {'width': '100%!IMPORTANT'}}, 'only', 'screen',
                              {"and": [{'max-width': '600px'}]})
    return st

  def arrow(self, records, width=(100, '%'), height=(70, 'px'), color=None, options=None):
    dft_options = {"shape": 'arrow'}
    if options is not None:
      dft_options.update(options)
    return self.stepper(records, width, height, color, dft_options)

  def rectangle(self, records, width=(100, '%'), height=(70, 'px'), color=None, options=None):
    dft_options = {"shape": 'rectangle'}
    if options is not None:
      dft_options.update(options)
    return self.stepper(records, width, height, color, dft_options)

  def triangle(self, records, width=(100, '%'), height=(70, 'px'), color=None, options=None):
    dft_options = {"shape": 'rectangle'}
    if options is not None:
      dft_options.update(options)
    return self.stepper(records, width, height, color, dft_options)

  def vertical(self, records, shape='circle',  width=(100, '%'), height=(70, 'px'), color=None, options=None):
    """
    Description:

    Entry point for the stepper object

    Usage::

      rptObj.ui.stepper([
        {"value": 'test 1', "status": 'success', 'label': 'test'},
        {"value": 'test 2', "status": 'error'},
        {"value": 'test 3', "status": 'pending'}])

    Attributes:
    ----------
    :param records: A list with the different steps defined in the workflow
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param color: Optional.
    :param options: Optional. Specific Python options available for this component
    """
    dft_options = {"line": False, "shape": shape}
    if options is not None:
      dft_options.update(options)
    st = html.HtmlStepper.Stepper(self.context.rptObj, records, width, height, color, dft_options)
    return st




from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsObject


class JsDOMRect(object):

  def __init__(self, varId):
    self.varId = varId

  def setVar(self, varName, varType="var"):
    return JsObject.JsObject("%s %s = %s" % (varType, varName, self.varId))

  @property
  def x(self):
    """
    Description:
    -----------
    The x read-only property of the DOMRectReadOnly interface represents the x coordinate of the DOMRect's origin.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/x
    """
    return JsNumber.JsNumber("%s.x" % self.varId)

  @property
  def y(self):
    """
    Description:
    -----------
    The y read-only property of the DOMRectReadOnly interface represents the y coordinate of the DOMRect's origin.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/y
    """
    return JsNumber.JsNumber("%s.y" % self.varId)

  @property
  def width(self):
    """
    Description:
    -----------
    The width read-only property of the DOMRectReadOnly interface represents the width of the DOMRect.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/width
    """
    return JsNumber.JsNumber("%s.width" % self.varId)

  @property
  def height(self):
    """
    Description:
    -----------
    The height read-only property of the DOMRectReadOnly interface represents the height of the DOMRect.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/height
    """
    return JsNumber.JsNumber("%s.height" % self.varId)

  @property
  def top(self):
    """
    Description:
    -----------
    The top read-only property of the DOMRectReadOnly interface returns the top coordinate value of the DOMRect.
    (Has the same value as y, or y + height if height is negative.)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/top
    """
    return JsNumber.JsNumber("%s.top" % self.varId)

  @property
  def right(self):
    """
    Description:
    -----------
    The right read-only property of the DOMRectReadOnly interface returns the right coordinate value of the DOMRect.
    (Has the same value as x + width, or x if width is negative.)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/right
    """
    return JsNumber.JsNumber("%s.top" % self.varId)

  @property
  def bottom(self):
    """
    Description:
    -----------
    The bottom read-only property of the DOMRectReadOnly interface returns the bottom coordinate value of the DOMRect.
    (Has the same value as y + height, or y if height is negative.)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/bottom
    """
    return JsNumber.JsNumber("%s.top" % self.varId)

  @property
  def left(self):
    """
    Description:
    -----------
    The left read-only property of the DOMRectReadOnly interface returns the left coordinate value of the DOMRect.
    (Has the same value as x, or x + width if width is negative.)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/DOMRectReadOnly/left
    """
    return JsNumber.JsNumber("%s.left" % self.varId)



from . import GraphC3
from . import GraphNVD3
from . import GraphChartJs
from . import GraphVis
from . import GraphSvg
from . import GraphBillboard
from . import GraphDC
from . import GraphD3
from . import GraphPlotly
from . import GraphSparklines
from . import GraphCanvas
from . import GraphGoogle




from epyk.core.html.options import Options


class OptionsJsonFmt(Options):

  @property
  def open(self):
    """
    Description:
    ------------
    Default: 1 This number indicates up to how many levels the rendered tree should expand.
    Set it to 0 to make the whole tree collapsed or set it to Infinity to expand the tree deeply

    Related Pages:
"""
    return self._config_get(1)

  @open.setter
  def open(self, bool):
    self._config(bool)

  @property
  def hoverPreviewEnabled(self):
    """
    Description:
    ------------
    Enable preview on hover.

    Related Pages:

      https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param bool:
    """
    return self._config_group_get('opts', False)

  @hoverPreviewEnabled.setter
  def hoverPreviewEnabled(self, bool):
    self._config_group('opts', bool)

  @property
  def hoverPreviewArrayCount(self):
    """
    Description:
    ------------
    Number of array items to show in preview Any array larger than this number will be shown as Array[XXX] where XXX is length of the array.

    Related Pages:

      https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param num:
    """
    return self._config_group_get('opts', 100)

  @hoverPreviewArrayCount.setter
  def hoverPreviewArrayCount(self, num):
    self._config_group('opts', num)

  @property
  def hoverPreviewFieldCount(self):
    """
    Description:
    ------------
    Number of object properties to show for object preview. Any object with more properties that thin number will be truncated.

    Related Pages:

      https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param num:
    """
    return self._config_group_get('opts', 5)

  @hoverPreviewFieldCount.setter
  def hoverPreviewFieldCount(self, num):
    self._config_group('opts', num)

  @property
  def animateOpen(self):
    """
    Description:
    ------------
    Enable animation when expanding json object. True by default.

    Related Pages:

      https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param bool:
    """
    return self._config_group_get('opts', True)

  @animateOpen.setter
  def animateOpen(self, bool):
    self._config_group('opts', bool)

  @property
  def animateClose(self):
    """
    Description:
    ------------
    Enable animation when closing json object. True by default.

    Related Pages:

      https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param bool:
    """
    return self._config_group_get('opts', True)

  @animateClose.setter
  def animateClose(self, bool):
    self._config_group('opts', bool)

  @property
  def useToJSON(self):
    """
    Description:
    ------------
    use the toJSON method to render an object as a string as available.
    Usefull for objects like Date or Mongo's ObjectID that migh make more sense as a strign than as empty objects. True by default.

    Related Pages:

      https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param bool:
    """
    return self._config_group_get('opts', True)

  @useToJSON.setter
  def useToJSON(self, bool):
    self._config_group('opts', bool)

  @property
  def sortPropertiesBy(self):
    """
    Description:
    ------------
    use the given sorting function to deeply sort the object properties.

    Related Pages:

      https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param bool:
    """
    return self._config_group_get('opts', None)

  @sortPropertiesBy.setter
  def sortPropertiesBy(self, bool):
    self._config_group('opts', bool)




from epyk.core.data.DataClass import DataClass

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects


class DragOptions(DataClass):

  @property
  def showTooltip(self):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/chrispahm/chartjs-plugin-dragdata
    """
    return self.get(True)

  @showTooltip.setter
  def showTooltip(self, bool):
    self.set(bool)


class DragData(DataClass):

  @property
  def dragX(self):
    """
    Description:
    -----------
    """
    return self.get(False)

  @dragX.setter
  def dragX(self, bool):
    self.set(bool)

  @property
  def dragDataRound(self):
    """
    Description:
    -----------
    """
    return self.get(0)

  @dragDataRound.setter
  def dragDataRound(self, num):
    self.set(num)

  @property
  def dragOptions(self):
    """
    Description:
    -----------
    """
    return self.sub_data("dragOptions", DragOptions)

  def onDragStart(self, jsFncs):
    """
    Description:
    -----------
    Called before zooming, return false to prevent the zoom

    Related Pages:

      https://github.com/chrispahm/chartjs-plugin-dragdata

    Attributes:
    ----------
    :param jsFncs:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onDragStart"] = JsObjects.JsVoid("function(event, element) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))

  def onDrag(self, jsFncs):
    """
    Description:
    -----------
    Change cursor style to grabbing during drag action

    Related Pages:

      https://github.com/chrispahm/chartjs-plugin-dragdata

    Attributes:
    ----------
    :param jsFncs:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onDrag"] = JsObjects.JsVoid("function(event, datasetIndex, index, value) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))

  def onDragEnd(self, jsFncs):
    """
    Description:
    -----------
    Restore default cursor style upon drag release

    Related Pages:

      https://github.com/chrispahm/chartjs-plugin-dragdata

    Attributes:
    ----------
    :param jsFncs:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onDragEnd"] = JsObjects.JsVoid("function(event, datasetIndex, index, value) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))




from epyk.core.js import JsUtils
from epyk.core.js import Js


f = JsUtils.JsFile(path=r"../outs")
js_obj = Js.JsBase()
breadcrumb = js_obj.breadcrumb
js_obj.addOnLoad(breadcrumb.add("test1", "value1"))
js_obj.addOnLoad(breadcrumb.add("test2", "value2"))
js_obj.addOnLoad(breadcrumb.hash("test3"))
js_obj.addOnLoad(js_obj.clipboard(breadcrumb.url))
js_obj.addOnLoad(js_obj.info("icon"))
f.codepen(js_obj)




import importlib, subprocess, sys, os, inspect

class EpykMissingPykException(Exception):
  pass

class EpykBoundRptObj(Exception):
  pass


def requires(pyk_file, autoinstall=None, autoreload=False):
  """
  Description:
    ------------

      This function will allow you to import components from another pyk file.
      The pyk file can be located on your file system, or it can be on pypi or even github
      if it's the latter options (pypi or github) you will need to use autoinstall=True

  Usage:
  -----------
    requires(pyk_name, autoinstall=True)

    requires('/usr/local/my_pyk_file.py')

  :param pyk_file: the pyk to be imported into your project, this will be a name if installing from pypi or github or just a path if it's a local pyk)
  :param components: list of objects to import, if None all the objects declared in the pyk file will be available
  :param autoinstall: specify whether we need to check github or pypi
  :return: a pyk object
  """

  pyk_obj = _Pyk.instance()
  return pyk_obj._requires(pyk_file, autoinstall, autoreload)

def exports(obj_dict):
  """
  Description:
  ------------

    This function requries you to pass a dictionary with the object alias as key and the object as value
    for as many objects as you wish
    This will then be used by users who need to require your particular pyk file
    You can pass epyk object in that dictionary as well as function if you need to

  Usage:
  ------

    exports({'my_obj1': my_obj1, 'my_obj1': my_obj2})

  :param obj__dict: dictionary with object name as key and object as value
  """
  pyk_obj = _Pyk.instance()
  pyk_obj._exports(obj_dict)


def register(rpt_obj, components):
  if type(components) != list:
    components = [components]

  for comp in components:
    if comp.htmlCode in rpt_obj.components:
      raise Exception("Duplicated Html Code %s in the script !" % comp.htmlCode)

    rpt_obj.components[comp.htmlCode] = comp


class _Pyk(object):
  """

  """
  __instance = None
  __pyk_dict = {}

  class Pyk(object):
    pass

  @staticmethod
  def instance():
    if _Pyk.__instance is None:
      _Pyk()
    return _Pyk.__instance

  def __init__(self):
    if _Pyk.__instance is not None:
      raise Exception("You need to call this object with the instance() method: _Pyk.instance()")

    _Pyk.__instance = self

  def _requires(self, pyk_file, autoinstall=False, autoreload=False):
    """
    Description:
    ------------

      This function will allow you to import components from another pyk file.
      The pyk file can be located on your file system, or it can be on pypi or even github
      if it's the latter options (pypi or github) you will need to use autoinstall=True

    Usage:
    -----------
      requires(pyk_name, autoinstall=True)

    :param pyk_file: the pyk to be imported into your project, this will be a name if installing from pypi or github or just a path if it's a local pyk)
    :param components: list of objects to import, if None all the objects declared in the pyk file will be available
    :param autoinstall: specify whether we need to check github or pypi
    :return: a pyk object
    """
    pyk_dict = _Pyk.__pyk_dict
    if pyk_file in pyk_dict and not autoreload:
      return pyk_dict[pyk_file]

    if pyk_file.endswith('.py'):
      path_split = pyk_file.split(os.sep)
      sys.path.append(os.sep.join(path_split[:-1]))
      importlib.import_module(path_split[-1].replace('.py', ''))
    else:
      try:
        importlib.import_module(pyk_file)
      except (ModuleNotFoundError, ImportError):
        if not autoinstall:
          print("The package is not found on your environment, you will need to install it manually or use the autoinstall argument")
          raise

        subprocess.check_call([sys.executable, '-m', 'pip', 'install', pyk_file])
        importlib.import_module((pyk_file))

    if pyk_file not in pyk_dict:
      raise EpykMissingPykException('The specified pyk file: %s does not call the exports function, if this is your file make sure to call this!' % pyk_file)

    return pyk_dict[pyk_file]

  def _exports(self, obj_dict):
    """
    Description:
    ------------

      This function requries you to pass a dictionary with the object alias as key and the object as value
      for as many objects as you wish
      This will then be used by users who need to require your particular pyk file
      You can pass epyk object in that dictionary as well as function if you need to

    Usage:
    ------

      exports({'my_obj1': my_obj1, 'my_obj1': my_obj2})

    :param obj__dict: dictionary with object name as key and object as value
    """
    pyk_dict = _Pyk.__pyk_dict
    pyk_obj = self.Pyk()
    frame_found = False
    for frame in inspect.stack():
      if frame_found:
        filename = frame.filename
        break

      if frame.function == 'exports':
        frame_found = True
    pyk_dict[filename] = {}
    for obj_name, obj in obj_dict.items():
      setattr(pyk_obj, obj_name, obj)
      pyk_dict[filename] = pyk_obj




"""
Wrapper to the Javascript Window module

Allows to save key/value pairs in a web browser. Stores the data with no expiration date
The localStorage and sessionStorage properties allow to save key/value pairs in a web browser.

Related Pages:

		https://www.w3schools.com/Jsref/prop_win_localstorage.asp
"""

from epyk.core.js import JsUtils
from epyk.core.js.objects import JsNodeDom

# All the predefined variable types
from epyk.core.js.fncs import JsFncs
from epyk.core.js.primitives import JsBoolean
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsObject


class JsLocalStorage(object):
  def key(self, i):
    """
    Description:
    ------------
    The key() method returns name of the key with the specified index.

    The localStorage object stores data with no expiration date.
    The data will not be deleted when the browser is closed, and will be available the next day, week, or year.

    Usage::

      jsObj.localStorage.key(0)

    Related Pages:

			https://www.w3schools.com/jsref/met_storage_key.asp

    Attributes:
    ----------
    :param i: Required. A Number representing the index of the key you want to get the name of

    :return: A String, representing the name of the specified key
    """
    i = JsUtils.jsConvertData(i, None)
    return JsObject.JsObject("localStorage.key(%s)" % i)

  def __setitem__(self, key, data):
    """
    Description:
    ------------

    Usage::

      :param key:
    :param data:

    :return:
    """
    self.setItem(key, data)

  def __getitem__(self, item):
    """
    Description:
    ------------
    Python wrapper to getItem

    Usage::

      jsObj.console.log(jsObj.sessionStorage["lastname"])

    Attributes:
    ----------
    :param item: String. The item name
    """
    return self.getItem(item)

  def setItem(self, key, data):
    """
    Description:
    ------------
    Syntax for SAVING data to localStorage

    The localStorage object stores data with no expiration date.
    The data will not be deleted when the browser is closed, and will be available the next day, week, or year.

    Usage::

      jsObj.localStorage.getItem("lastname", "test")

    Related Pages:

			https://www.w3schools.com/jsref/met_storage_setitem.asp

    Attributes:
    ----------
    :param key: Required. A String specifying the name of the key you want to set the value of
    :param data: Required. A String specifying the value of the key you want to set the value of

    :return: A String, representing the inserted value
    """
    key = JsUtils.jsConvertData(key, None)
    data = JsUtils.jsConvertData(data, None)
    return JsObject.JsObject("localStorage.setItem(%s, %s)" % (key, data))

  def getItem(self, key):
    """
    Description:
    ------------
    Syntax for READING data from localStorage:

    The localStorage object stores data with no expiration date.
    The data will not be deleted when the browser is closed, and will be available the next day, week, or year.

    Usage::

      jsObj.localStorage.getItem("lastname")

    Related Pages:

			https://www.w3schools.com/jsref/met_storage_getitem.asp

    Attributes:
    ----------
    :param key: Required. A String specifying the name of the key you want to get the value of

    :return: A String, representing the value of the specified key
    """
    key = JsUtils.jsConvertData(key, None)
    return JsObject.JsObject("localStorage.getItem(%s)" % key)

  def removeItem(self, key):
    """
    Description:
    ------------
    The removeItem() method removes the specified Storage Object item.

    The localStorage object stores data with no expiration date.
    The data will not be deleted when the browser is closed, and will be available the next day, week, or year.

    Usage::

      jsObj.localStorage.removeItem("lastname")

    Related Pages:

			https://www.w3schools.com/jsref/met_storage_removeitem.asp

    Attributes:
    ----------
    :param key: Required. A String specifying the name of the item you want to remove

    :return: Void
    """
    key = JsUtils.jsConvertData(key, None)
    return JsFncs.JsFunction("localStorage.removeItem(%s)" % key)

  def clear(self):
    """
    Description:
    ------------
    The clear() method removes all the Storage Object item for this domain.

    The localStorage object stores data with no expiration date.
    The data will not be deleted when the browser is closed, and will be available the next day, week, or year.

    Usage::

      jsObj.localStorage.clear()

    Related Pages:

			https://www.w3schools.com/jsref/met_storage_clear.asp

    :return: Void
    """
    return JsFncs.JsFunction("localStorage.clear()")


class JsSessionStorage(object):
  """
  Description:
  ------------
  The localStorage and sessionStorage properties allow to save key/value pairs in a web browser.

  The sessionStorage object stores data for only one session (the data is deleted when the browser tab is closed).

  Related Pages:
  --------------
  https://www.w3schools.com/Jsref/prop_win_sessionstorage.asp
  """

  def key(self, i):
    """
    Description:
    ------------
    The sessionStorage object stores data for only one session (the data is deleted when the browser tab is closed).

    Attributes:
    ----------
    :param i: The key number
    """
    i = JsUtils.jsConvertData(i, None)
    return JsObject.JsObject("sessionStorage.key(%s)" % i)

  def __setitem__(self, key, data):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param key:
    :param data:
    """
    self.setItem(key, data)

  def __getitem__(self, item):
    """
    Description:
    ------------
    Python wrapper to getItem

    Usage::

      jsObj.console.log(jsObj.sessionStorage["lastname"])

    Attributes:
    ----------
    :param item:
    """
    return self.getItem(item)

  def setItem(self, key, data):
    """
    Description:
    ------------
    Syntax for SAVING data to sessionStorage

    The sessionStorage object stores data for only one session (the data is deleted when the browser tab is closed).

    Usage::

      jsObj.sessionStorage.setItem("lastname", "Smith")
      jsObj.sessionStorage.setItem("lastname", jsObj.objects.get("bin")),

    Related Pages:

			https://www.w3schools.com/Jsref/prop_win_sessionstorage.asp

    Attributes:
    ----------
    :param key: The key used to store the data in the session cache
    :param data:
    """
    key = JsUtils.jsConvertData(key, None)
    data = JsUtils.jsConvertData(data, None)
    return JsObject.JsObject("sessionStorage.setItem(%s, %s)" % (key, data))

  def getItem(self, key):
    """
    Description:
    ------------
    Syntax for READING data from sessionStorage

    The sessionStorage object stores data for only one session (the data is deleted when the browser tab is closed).

    Usage::

      jsObj.sessionStorage.getItem("lastname")
      jsObj.console.log(jsObj.sessionStorage.getItem("lastname"))

    Attributes:
    ----------
    :param key:
    """
    key = JsUtils.jsConvertData(key, None)
    return JsObject.JsObject("sessionStorage.getItem(%s)" % key)

  def removeItem(self, jsData, jsDataKey=None, isPyData=False, jsFnc=None):
    """
    Description:
    ------------
    Syntax for REMOVING ALL saved data from sessionStorage

    The sessionStorage object stores data for only one session (the data is deleted when the browser tab is closed).

    Usage::

      jsObj.sessionStorage.removeItem("lastname")

    Related Pages:

			https://www.w3schools.com/jsref/met_storage_removeitem.asp

    Attributes:
    ----------
    :param jsData:
    :param jsDataKey:
    :param isPyData:
    :param jsFnc:
    """
    jsData = JsUtils.jsConvert(jsData, jsDataKey, isPyData, jsFnc)
    return JsFncs.JsFunction("sessionStorage.removeItem(%s)" % jsData)

  def clear(self):
    """
    Description:
    ------------
    Syntax for REMOVING ALL saved data from sessionStorage

    The sessionStorage object stores data for only one session (the data is deleted when the browser tab is closed).

    Related Pages:

			https://www.w3schools.com/jsref/prop_win_sessionstorage.asp
    """
    return JsFncs.JsFunction("sessionStorage.clear()")


class JsHistory(object):
  """
  Description:
  ------------
  Interface to the Javascript history module

  Related Pages:
  --------------
  https://www.w3schools.com/js/js_window_history.asp
  """

  def __init__(self, context):
    self.__jsObj = context

  @property
  def length(self):
    """
    Description:
    ------------
    The length property returns the number of URLs in the history list of the current browser window.

    Usage::

      rptObj.js.window.history.length

    Related Pages:

			https://www.w3schools.com/jsref/prop_his_length.asp

    :return: A Number, representing the number of entries in the session history
    """
    return JsNumber.JsNumber("history.length", isPyData=False)

  def back(self):
    """
    Description:
    ------------
    The back() method loads the previous URL in the history list.

    Usage::

      rptObj.js.window.history.back()

    Related Pages:

			https://www.w3schools.com/jsref/met_his_back.asp

    :return: The Javascript String to be added to the page
    """
    return JsFncs.JsFunction("window.history.back()")

  def forward(self):
    """
    Description:
    ------------
    The forward() method loads the next URL in the history list.

    Related Pages:

			https://www.w3schools.com/jsref/met_his_forward.asp

    :return: The Javascript String to be added to the page
    """
    return JsFncs.JsFunction("window.history.forward()")

  def go(self, number):
    """
    Description:
    ------------
    The go() method loads a specific URL from the history list.

    Related Pages:

			https://www.w3schools.com/jsref/met_his_go.asp

    Attributes:
    ----------
    :param number: Required. The parameter can either be a number which goes to the URL within the specific position (-1 goes back one page, 1 goes forward one page), or a string.

    :return: The Javascript String to be added to the page
    """
    return JsFncs.JsFunction("window.history.go(%s)" % number)

  def pushState(self, stateObj, title, url):
    """
    Description:
    ------------
    Pushes the given data onto the session history stack with the specified title and, if provided, URL

    Note that pushState() never causes a hashchange event to be fired, even if the new URL differs from the old URL only in its hash

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/History_API

    Attributes:
    ----------
    :param stateObj: he state object is a JavaScript object which is associated with the new history entry created by pushState()
    :param title: Firefox currently ignores this parameter, although it may use it in the future
                  Passing the empty string here should be safe against future changes to the method.
                  Alternatively, you could pass a short title for the state to which you're moving.
    :param url: The new history entry's URL is given by this parameter.
                Note that the browser won't attempt to load this URL after a call to pushState(),

    :return:
    """
    return JsFncs.JsFunction("window.history.pushState('%s', '%s', %s)" % (stateObj, title, url))

  def replaceState(self, stateObj, title, url):
    """
    Description:
    ------------
    history.replaceState() operates exactly like history.pushState() except that replaceState() modifies the current history entry instead of creating a new one.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/History_API

    Attributes:
    ----------
    :param stateObj: he state object is a JavaScript object which is associated with the new history entry created by pushState()
    :param title: Firefox currently ignores this parameter, although it may use it in the future
                  Passing the empty string here should be safe against future changes to the method.
                  Alternatively, you could pass a short title for the state to which you're moving.
    :param url: The new history entry's URL is given by this parameter.
                Note that the browser won't attempt to load this URL after a call to pushState(),
    """
    return JsFncs.JsFunction("window.history.replaceState('%s', '%s', %s)" % (stateObj, title, url))

  def updateState(self, key, val):
    """
    Description:
    ------------
    Wrapper function

    This function is a simple wrapping function on top of the pushState history method.
    The purpose of this method is to make easier the update of the url whenever a component in the framework is updated.

    Usage::

      htmlObj.js.js.window.history.updateState(self.htmlCode, self.val)

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/History_API

    Attributes:
    ----------
    :param key: The key to be added or updated in the current URL
    :param val: The value to be changed to the current URL

    :return: The Javascript String for the method
    """
    self.__jsObj.registerFunction("updateUrl", [
      self.__jsObj.objects.new([], varName="newPmts"),
      self.__jsObj.location.search.substr(1).split("&").forEach([
        self.__jsObj.if_(self.__jsObj.data.loop.val.toString(explicit=False).includes("=", jsObj=self.__jsObj), [
          self.__jsObj.objects.array.new(self.__jsObj.data.loop.val.toString().split("="), varName="urlPmts"),
          self.__jsObj.objects.array.get("urlPmts")[0].toString(),
          self.__jsObj.objects.get("newPmts").addItem(self.__jsObj.objects.array.get("urlPmts")[0], self.__jsObj.objects.array.get("urlPmts")[1])
        ])
      ]),

      self.__jsObj.objects.get("newPmts").addItem(self.__jsObj.objects.get("urlKey"), self.__jsObj.objects.get("urlValue")),
      # Then we concatenate the URL
      self.__jsObj.objects.array.new([], varName="pmts"),
      self.__jsObj.objects.get("newPmts").entries().forEach([
        self.__jsObj.objects.array.get("pmts").push(self.__jsObj.data.loop.val[0].toString(explicit=False).add("=").add(self.__jsObj.data.loop.val[1]))
      ]),
      self.__jsObj.return_(self.__jsObj.location.origin + self.__jsObj.location.pathname + "?" + self.__jsObj.objects.array.get("pmts").join("&"))
      ], ["urlKey", "urlValue"])
    return self.pushState("data", "", JsFncs.JsFunction("updateUrl(%s, %s)" % (JsUtils.jsConvertData(key, None), JsUtils.jsConvertData(val, None))))


class JsWindowEvent(object):
  def addEventListener(self, eventType, jsFncs, windowId="window", subEvents=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param eventType:
    :param jsFncs:
    :param windowId:
    :param subEvents: List of names you want your underlying function to have as arguments
    """
    subEvents = '' if not subEvents else ','.join(subEvents)
    eventType = JsUtils.jsConvertData(eventType, None)
    jsFncs = JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return JsFncs.JsFunction("%s.addEventListener(%s, function(%s){%s})" % (windowId, eventType, subEvents, jsFncs))

  def addScrollListener(self, jsFncs, windowId="window"):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs:
    :param windowId:
    """
    return self.addEventListener("scroll", jsFncs, windowId)

  def addContentLoaded(self, jsFncs, windowId="window"):
    """
    Description:
    ------------
    The DOMContentLoaded event fires when the initial HTML document has been completely loaded and parsed, without waiting for stylesheets, images, and subframes to finish loading.

    Usage::

      rptObj.js.addOnLoad(
      rptObj.js.window.events.addContentLoaded(rptObj.js.alert("DOM fully loaded and parsed"))
      )

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Window/DOMContentLoaded_event

    Attributes:
    ----------
    :param jsFncs:
    :param windowId:
    """
    return self.addEventListener("DOMContentLoaded", jsFncs, windowId)

  def addClickListener(self, jsFncs, windowId="window", subEvents=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs:
    :param windowId:
    """
    return self.addEventListener("click", jsFncs, windowId, subEvents)


class JsWindow(object):
  """
  Description:
  ------------
  The window object represents an open window in a browser.

  If a document contain frames (<iframe> tags), the browser creates one window object for the HTML document,
  and one additional window object for each frame.

  Related Pages:
  --------------
  https://www.w3schools.com/Jsref/obj_window.asp
  """
  @property
  def scrollY(self, windowId="window"):
    """
    Description:
    ------------
    The read-only scrollY property of the Window interface returns the number of pixels that the document is currently scrolled vertically.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Window/scrollY

    Attributes:
    ----------
    :param windowId: String. The window reference
    """
    return JsNumber.JsNumber("%s.scrollY" % windowId)

  @property
  def innerHeight(self, windowId="window"):
    """
    Description:
    ------------
    The innerHeight property returns the height of a window's content area.

    Related Pages:

			https://www.w3schools.com/jsref/prop_win_innerheight.asp

    Attributes:
    ----------
    :param windowId: String. The window reference
    """
    return JsNumber.JsNumber("%s.innerHeight" % windowId)

  @property
  def scrollEndPage(self, windowId="window"):
    """
    Description:
    ------------
    The scrollEndPage property indicates if the page is scrolled to the end

    Attributes:
    ----------
    :param windowId: String. The window reference
    """
    return JsBoolean.JsBoolean("(%s.scrollY + %s.innerHeight > document.body.clientHeight)? true: false" % (windowId, windowId), isPyData=False)

  @property
  def scrollPercentage(self, windowId="window"):
    """
    Description:
    ------------
    The scrollPercentage property return the percentage of the page scrolled

    Attributes:
    ----------
    :param windowId: String. The window reference
    """
    return JsNumber.JsNumber(
      "Math.min((%s.scrollY + %s.innerHeight)/ document.body.clientHeight * 100, 100)" % (windowId, windowId), isPyData=False)

  @property
  def scrollMaxY(self, windowId="window"):
    """
    Description:
    ------------
    The Window.scrollMaxY read-only property returns the maximum number of pixels that the document can be scrolled vertically.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Window/scrollMaxY
    """
    return JsNumber.JsNumber("(%s.scrollMaxY || (document.documentElement.scrollHeight - document.documentElement.clientHeight))" % windowId)

  def __init__(self, src=None):
    """
    Description:
    ------------

    :type src: epyk.Lib.js.Js.JsBase

    Attributes:
    ----------
    :param src:
    """
    self.__src = src
    self._context = src._context if hasattr(src, '_context') else {}

  @property
  def document(self):
    """
    Description:
    ------------
    Interface to the DOM object on the current window

    :return: A Python JsDoms object wrapping the DOM Js interface
    """
    return JsNodeDom.JsDoms(self.__src)

  @property
  def history(self):
    """
    Description:
    ------------
    Interface to the History object

    :return: A Python Js History object
    """
    return JsHistory(self.__src)

  def close(self, windowId="window"):
    """
    Description:
    ------------
    Closes the current window

    Related Pages:

			https://www.w3schools.com/jsref/met_win_close.asp

    Attributes:
    ----------
    :param windowId: The JavaScript window object

    :return: The String representing the Javascript function
    """
    return JsFncs.JsFunction("%s.close()" % windowId)

  @property
  def events(self):
    """
    Description:
    ------------
    Property to all the events
    """
    return JsWindowEvent()

  def addEventListener(self, eventType, jsFncs, windowId="window"):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param eventType:
    :param jsFncs:
    :param windowId:
    """
    eventType = JsUtils.jsConvertData(eventType, None)
    jsFncs = JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return JsFncs.JsFunction("%s.addEventListener(%s, function(){%s})" % (windowId, eventType, jsFncs))

  def open(self, url, name="_self", specs=None, replace=None, windowId="window"):
    """
    Description:
    ------------
    Opens a new browser window

    Related Pages:

			https://www.w3schools.com/Jsref/met_win_open.asp

    Attributes:
    ----------
    :param url: Optional. Specifies the URL of the page to open. If no URL is specified, a new window/tab with about:blank is opened
    :param name: Optional. Specifies the target attribute or the name of the window.
    :param specs: Optional. A comma-separated list of items, no whitespaces.
    :param replace: Optional. Specifies whether the URL creates a new entry or replaces the current entry in the history list
    :param windowId: The JavaScript window object
    """
    url = JsUtils.jsConvertData(url, None)
    name = JsUtils.jsConvertData(name, None)
    specs = JsUtils.jsConvertData(specs, None)
    replace = JsUtils.jsConvertData(replace, None)
    return JsFncs.JsFunction("%s.open(%s, %s, %s, %s)" % (windowId, url, name, specs, replace))

  def postData(self, jsData):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsData:
    """

    if not isinstance(jsData, list):
      jsData = [jsData]

  def download(self, data, fileName):
    """
    Description:
    ------------
    Download the data from a flat file

    Usage::

      rptObj.js.window.download(rptObj.js.window.btoa(rptObj.js.objects.get("test")), fileName="test.txt")

    Attributes:
    ----------
    :param data:
    :param fileName:

    :return: Void,
    """
    data = JsUtils.jsConvertData(data, None)
    return JsFncs.JsFunction(";".join(JsUtils.jsConvertFncs([
      self.__src.createElement("a", varName="a_temp").setAttribute("download", fileName).setAttribute("href", JsFncs.JsFunction("'data:text/csv;base64,'+ %s" % data)),
      self.__src.body.appendChild(self.__src.objects.get("a_temp")),
      self.__src.objects.dom.get("a_temp").click(),
      #self.__src.objects.dom.get("a_temp").remove()
    ])))
    #return JsFncs.JsFunction("%s.open('data:text/csv;base64,'+ %s, '_self')" % (windowId, data))

  def moveBy(self, x, y, windowId="window"):
    """
    Description:
    ------------
    The moveBy() method moves a window a specified number of pixels relative to its current coordinates.

    Related Pages:

			https://www.w3schools.com/Jsref/met_win_moveby.asp

    Attributes:
    ----------
    :param x: The horizontal move in pixel
    :param y: The vertical move in pixel
    :param windowId: The JavaScript window object
    """
    x = JsUtils.jsConvertData(x, None)
    y = JsUtils.jsConvertData(y, None)
    return JsFncs.JsFunction("%s.moveBy(%s, %s)" % (windowId, x, y))

  def focus(self, windowId="window"):
    """
    Description:
    ------------
    The focus() method sets focus to the current window

    Related Pages:

			https://www.w3schools.com/Jsref/met_win_focus.asp

    Attributes:
    ----------
    :param windowId: The JavaScript window object

    :return: Void, The Javascript String
    """
    return JsFncs.JsFunction("%s.focus()" % windowId)

  def scroll(self, x, y, windowId="window"):
    """
    Description:
    ------------
    The Window.scroll() method scrolls the window to a particular place in the document.

    Related Pages:

			https://developer.mozilla.org/uk/docs/Web/API/Window/scroll

    Attributes:
    ----------
    :param x: The pixel along the horizontal axis of the document that you want displayed in the upper left
    :param y: The pixel along the vertical axis of the document that you want displayed in the upper left
    """
    return JsFncs.JsFunction("%s.scroll(%s, %s)" % (windowId, x, y))

  def scrollTo(self, x=None, y=None, windowId="window"):
    """
    Description:
    ------------
    The window.scrollTo() go to a particular point

    Attributes:
    ----------
    :param x:
    :param y:
    :param windowId:
    """
    y = y or "document.body.scrollHeight"
    x = x or "document.body.scrollWidth"
    return JsFncs.JsFunction("%s.scrollTo(%s, %s)" % (windowId, x, y))

  def scrollUp(self, windowId="window"):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param windowId:
    """
    return JsFncs.JsFunction("%s.scrollTo(0, 0)" % windowId)

  def print(self, windowId="window"):
    """
    Description:
    ------------
    Prints the content of the current window

    Related Pages:

			https://www.w3schools.com/Jsref/met_win_print.asp

    Attributes:
    ----------
    :param windowId: The JavaScript window object

    :return: Void, The Javascript String
    """
    return JsFncs.JsFunction("%s.print()" % windowId)

  def alert(self, jsData, jsFnc=None, windowId="window", skip_data_convert=False):
    """
    Description:
    ------------
    The alert() method displays an alert box with a specified message and an OK button.

    Usage::

      rptObj.js.window.alert("Test")
      rptObj.js.alert("Test 2")

    Related Pages:

			https://www.w3schools.com/jsref/met_win_alert.asp

    Attributes:
    ----------
    :param jsData: Optional. Specifies the text to display in the alert box, or an object converted into a string and displayed
    :param jsFnc: A JsFnc or a list of JsFncs
    :param windowId: The JavaScript window object
    """
    if skip_data_convert:
      return JsFncs.JsFunction("%s.alert(%s)" % (windowId, jsData))

    return JsFncs.JsFunction("%s.alert(%s)" % (windowId, JsUtils.jsConvertData(jsData, jsFnc)))

  def atob(self, jsData, jsFnc=None, windowId="window"):
    """
    Description:
    ------------
    Decodes a base-64 encoded string

    Usage::

      jsObj.window.btoa("Test").setVar("bin")
      jsObj.window.atob(jsObj.objects.get("bin"))

    Related Pages:

			https://www.w3schools.com/jsref/met_win_atob.asp

    Attributes:
    ----------
    :param jsData: Required. The string which has been encoded by the btoa() method
    :param jsFnc: A JsFnc or a list of JsFncs
    :param windowId: The JavaScript window object
    """
    return JsFncs.JsFunction("%s.atob(%s)" % (windowId, JsUtils.jsConvertData(jsData, jsFnc)))

  def btoa(self, jsData, jsFnc=None, windowId="window"):
    """
    Description:
    ------------
    Encodes a string in base-64

    Usage::

      jsObj.window.btoa("Test").setVar("bin")

    Related Pages:

			https://www.w3schools.com/jsref/met_win_btoa.asp

    Attributes:
    ----------
    :param jsData: Required. The string to be encoded
    :param jsFnc: A JsFnc or a list of JsFncs
    :param windowId: The JavaScript window object
    """
    return JsObject.JsObject("%s.btoa(%s)" % (windowId, JsUtils.jsConvertData(jsData, jsFnc)), isPyData=False)

  def setInterval(self, jsFncs, varId, milliseconds, windowId="window", setVar=True):
    """
    Description:
    ------------
    The setInterval() method calls a function or evaluates an expression at specified intervals (in milliseconds).

    The setInterval() method will continue calling the function until clearInterval() is called, or the window is closed.

    Usage::

      jsObj.window.setInterval([jsObj.console.log(jsObj.math.random())], 5000)

    Related Pages:

			https://www.w3schools.com/jsref/met_win_setinterval.asp

    #TODO: Add a control on setInterval to only have one created

    Attributes:
    ----------
    :param jsFncs: Required. The function that will be executed
    :param varId:
    :param milliseconds: Required. The intervals (in milliseconds) on how often to execute the code. If the value is less than 10, the value 10 is used
    :param windowId: The JavaScript window object
    :param setVar: Boolean.
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    if setVar:
      if windowId == 'window':
        # To make the variable scope as global
        varId = "window['%s']" % varId
      else:
        varId = "var %s" % varId

      return JsFncs.JsFunction("%s = %s.setInterval(function(){%s}, %s)" % (varId, windowId, ";".join(jsFncs), milliseconds))

    return JsFncs.JsFunction("%s.setInterval(function(){%s}, %s)" % (windowId, ";".join(jsFncs), milliseconds))

  def clearInterval(self, varId, windowId="window"):
    """
    Description:
    ------------
    The clearInterval() method clears a timer set with the setInterval() method.

    The ID value returned by setInterval() is used as the parameter for the clearInterval() method.

    Usage::

      jsObj.window.setInterval([jsObj.console.log(jsObj.math.random())], 500).setVar("interva1"),
      jsObj.window.clearInterval(jsObj.objects.get("interva1"))

    Related Pages:

			https://www.w3schools.com/jsref/met_win_clearinterval.asp

    #TODO: Check if interval is unique

    Attributes:
    ----------
    :param varId: A PythonJs object (JsArray, JsObject...)
    :param windowId: The JavaScript window object

    :return: Void, The Javascript String
    """
    js_data = varId if not hasattr(varId, "toStr") else JsUtils.jsConvertData(varId, None)
    return JsFncs.JsFunction("%s.clearInterval(%s); %s = undefined" % (windowId, js_data, js_data))

  def toggleInterval(self, jsFncs, varId, milliseconds, windowId="window"):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.button("Interval Toggle").click([
      rptObj.js.window.toggleInterval(rptObj.js.console.log('Print called'), 'test', 400),
    ])

    Attributes:
    ----------
    :param jsFncs:
    :param varId:
    :param milliseconds:
    :param windowId:
    """
    interval = self.setInterval(jsFncs, varId, milliseconds, windowId, setVar=False)
    clear = self.clearInterval(varId, windowId)
    return JsFncs.JsFunction("if(%s){%s = %s} else{%s}" % (JsUtils.isNotDefined(varId), varId, interval, clear))

  def setTimeout(self, jsFncs, milliseconds=0, windowId="window"):
    """
    Description:
    ------------
    The setTimeout() method calls a function or evaluates an expression after a specified number of milliseconds.

    Related Pages:

			https://www.w3schools.com/jsref/met_win_settimeout.asp

    Attributes:
    ----------
    :param jsFncs: Required. The function that will be executed
    :param milliseconds: 	Optional. The number of milliseconds to wait before executing the code. If omitted, the value 0 is used
    :param windowId: The JavaScript window object
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return JsObject.JsObject("%s.setTimeout(function(){%s}, %s)" % (windowId, ";".join(jsFncs), milliseconds), setVar=True, isPyData=False)

  def clearTimeout(self, jsData, jsFnc=None, windowId="window"):
    """
    Description:
    ------------
    The clearTimeout() method clears a timer set with the setTimeout() method.
    The ID value returned by setTimeout() is used as the parameter for the clearTimeout() method.

    Related Pages:

			https://www.w3schools.com/jsref/met_win_cleartimeout.asp

    Attributes:
    ----------
    :param windowId: The JavaScript window object
    """
    return JsFncs.JsFunction("%s.clearTimeout(%s)" % (windowId, JsUtils.jsConvertData(jsData, jsFnc)))

  def getComputedStyle(self, jsElement, jsPseudoElement=None, windowId="window"):
    """
    Description:
    ------------
    The getComputedStyle() method gets all the actual (computed) CSS property and values of the specified element.

    Related Pages:

			https://www.w3schools.com/jsref/jsref_getcomputedstyle.asp

    Attributes:
    ----------
    :param jsElement: Required. The element to get the computed style for
    :param windowId: The JavaScript window object

    :return: A CSSStyleDeclaration object containing CSS declaration block of the element
    """
    if jsPseudoElement is None:
      return JsFncs.JsFunction("%s.getComputedStyle(%s)" % (windowId, jsElement))

    return JsFncs.JsFunction("%s.getComputedStyle(%s, %s)" % (windowId, jsElement, jsPseudoElement))

  def getSelection(self, windowId="window"):
    """
    Description:
    ------------
    Returns a Selection object representing the range of text selected by the user

    Attributes:
    ----------
    :param windowId: The JavaScript window object
    """
    return JsFncs.JsFunction("%s.getSelection()" % windowId)

  def getVar(self, varName, windowId="window"):
    """
    Description:
    ------------
    Get the Javascript Variable name

    Attributes:
    ----------
    :param varName: The Variable name
    :param windowId: The JavaScript window object

    :return: Return the piece of script to be added to the Javascript
    """
    return "%s['%s']" % (windowId, varName)

  def onPageShow(self, jsFncs):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._context.setdefault('pageshow', []).extend(jsFncs)

  def onBeforeUnload(self, jsFncs):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs: A JsFnc or a list of JsFncs
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._context.setdefault('beforeunload', []).extend(jsFncs)




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects


class JsSwitch(object):
  """
    Description:
    ------------

  Documentation:
    - https://www.w3schools.com/js/js_switch.asp

  """

  def __init__(self, variable):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param variable:
    """
    self.__selector = variable
    self.__js, self.__default = [], None

  def case(self, value, jsFncs, strict=False):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param value: Object. The pivot value
    :param jsFncs: List or String. The JavaScript functions
    :param strict: Boolean.
    """
    if strict:
      self.__js.append((JsObjects.JsVoid("(%s === %s)" % (self.__selector, value)), jsFncs))
    else:
      self.__js.append((JsObjects.JsVoid("(%s == %s)" % (self.__selector, value)), jsFncs))
    return self

  def caseRange(self, min, max, jsFncs, include_value=True):
    """
    Description:
    ------------
    Range case. The value should be within the range [min, max]

    Attributes:
    ----------
    :param max: Number. The min value for this range
    :param min:  Number. The max value for this range
    :param jsFncs: List or String. The JavaScript functions
    :param include_value: Boolean. To specify if the pivot value is included
    """
    max = JsUtils.jsConvertData(max, None)
    min = JsUtils.jsConvertData(min, None)
    if include_value:
      self.__js.append((JsObjects.JsVoid("(%s <= %s) && (%s <= %s)" % (min, self.__selector, self.__selector, max)), jsFncs))
    else:
      self.__js.append((JsObjects.JsVoid("(%s < %s) && (%s < %s)" % (min, self.__selector, self.__selector, max)), jsFncs))
    return self

  def caseBelow(self, value, jsFncs, include_value=True):
    """
    Description:
    ------------
    Below case. The switch value should be below the value.

    Attributes:
    ----------
    :param value: Number. The pivot value
    :param jsFncs: List or String. The JavaScript functions
    :param include_value: Boolean. To specify if the pivot value is included
    """
    value = JsUtils.jsConvertData(value, None)
    if include_value:
      self.__js.append((JsObjects.JsVoid("(%s <= %s)" % (self.__selector, value)), jsFncs))
    else:
      self.__js.append((JsObjects.JsVoid("(%s < %s)" % (self.__selector, value)), jsFncs))
    return self

  def caseAbove(self, value, jsFncs, include_value=True):
    """
    Description:
    ------------
    Above case. The switch value should be above the value.

    Attributes:
    ----------
    :param value: Number. The pivot value
    :param jsFncs: List or String. The JavaScript functions
    :param include_value: Boolean. To specify if the pivot value is included
    """
    value = JsUtils.jsConvertData(value, None)
    if include_value:
      self.__js.append((JsObjects.JsVoid("(%s >= %s)" % (self.__selector, value)), jsFncs))
    else:
      self.__js.append((JsObjects.JsVoid("(%s > %s)" % (self.__selector, value)), jsFncs))
    return self

  def default_(self, jsFncs):
    """
    Description:
    ------------
    Default case value

    Attributes:
    ----------
    :param jsFncs: List or String. The JavaScript functions
    """
    self.__default = jsFncs
    return self

  def toStr(self):
    """
    Description:
    ------------

    """
    strData = []
    for var, jsFncs in self.__js:
      strData.append("case %s: {%s; break}" % (JsUtils.jsConvertData(var, None), JsUtils.jsConvertFncs(jsFncs, toStr=True)))
    if self.__default is not None:
      strData.append("default: {%s}" % JsUtils.jsConvertFncs(self.__default, toStr=True))
    self.__jsFncs, self.__default = [], None # empty the stack
    return "switch (true){%s}" % ("".join(strData))

  def __str__(self):
    return self.toStr()




from epyk.core.html.options import Options


class OptionsTree(Options):

  @property
  def is_root(self):
    """

    :return:
    """
    return self._config_get(True)

  @is_root.setter
  def is_root(self, bool):
    self._config(bool)

  @property
  def icon_open(self):
    """

    :return:
    """
    return self._config_get("fas fa-folder-open")

  @icon_open.setter
  def icon_open(self, icon):
    self._config(icon)

  @property
  def icon_close(self):
    """

    :return:
    """
    return self._config_get("fas fa-folder")

  @icon_close.setter
  def icon_close(self, icon):
    self._config(icon)

  @property
  def expanded(self):
    """

    :return:
    """
    return self._config_get(True)

  @expanded.setter
  def expanded(self, bool):
    self._config(bool)

  @property
  def style(self):
    """

    :return:
    """
    return self._config_get({})

  @style.setter
  def style(self, css):
    self._config(css)


class OptDropDown(Options):

  @property
  def width(self):
    """
    Description:
    ------------
    """
    return self._config_get(False)

  @width.setter
  def width(self, value):
    if isinstance(value, int):
      value = "%spx" % value
    #
    self._config_group_get("a", value)
    self._config_group_get("ul", value, name="left")




import json
import collections
import time

try:
  basestring
except NameError:
  basestring = str

from epyk.core.js import Imports
from epyk.interfaces import Components
from epyk.core.css.themes import Theme
from epyk.core.css import Classes

from epyk.core import html
from epyk.core import js
from epyk.core import data

from epyk.core.html import symboles
from epyk.core.html import entities
from epyk.core.py import OrderedSet
from epyk.core.py import PyOuts
from epyk.core.py import PyExt


class Report(object):
  showNavMenu, withContainer = False, False
  ext_packages = None # For extension modules

  def __init__(self, inputs=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param inputs: Dictionary. The global input data for the defined components in the page.
                               Passing data for a given component with an htmlCode will override the value.
    """
    self._css, self._ui, self._js, self._py, self._theme, self.__body = {}, None, None, None, None, None
    self._tags, self._header_obj, self.__import_manage = None, None, None
    self._props = {'js': {'onReady': OrderedSet(), 'events': {}, 'datasets': {}, 'configs': {}, 'builders': OrderedSet()},
                   'context': {'framework': 'JS'}}
    self.components = collections.OrderedDict() # Components for the entire page
    self.start_time, self.inputs, self._propagate = time.time(), inputs or {}, []
    self._scroll, self._contextMenu = set(), {}
    self.logo, self._dbSettings, self.dbsDef, self._cssText, self._jsText = None, None, {}, [], [] # to be reviewed

    self.jsImports, self.jsLocalImports, self.cssImport, self.cssLocalImports = set(), set(), set(), set()

  @property
  def body(self):
    """
    Description:
    ------------
    Property that returns the Body element of the HTML page
    """
    if self.__body is None:
      self.__body = html.Html.Body(self, None, htmlCode='body')
    return self.__body

  @body.setter
  def body(self, calc):
    self.__body = calc(self, None)

  @property
  def theme(self):
    """
    Description:
    ------------
    Return the currently used :doc:`report/theme` for the report
    """
    if self._theme is None:
      self._theme = Theme.ThemeDefault()
    return self._theme

  @theme.setter
  def theme(self, theme):
    if isinstance(theme, dict):
      self._theme = Theme.ThemeCustome()
      self._theme = theme
    else:
      self._theme = theme

  def imports(self, online=False):
    """
    Description:
    ------------
    Return the :doc:`report/import_manager`, which allows to import automatically packages for certain components to run.
    """
    if self.__import_manage is None:
      self.__import_manage = Imports.ImportManager(online, report=self)
    return self.__import_manage

  @property
  def symbols(self):
    """
    Description:
    ------------
    Shortcut to the HTML symbols

    Related Pages:

      https://www.w3schools.com/html/html_symbols.asp
      https://www.w3schools.com/charsets/ref_utf_math.asp

    Those can be added in string in order to improve the render of a text.
    """
    return symboles.Symboles()

  @property
  def entities(self):
    """
    Description:
    ------------
    Shortcut to the HTML Entities

    Related Pages:

      https://www.w3schools.com/html/html_entities.asp

    Those can be added in string in order to improve the render of a text.
    """
    return entities.Entities()

  @property
  def ui(self):
    """
    Description:
    ------------
    User Interface section.

    All the :doc:`components <report/ui>` which can be used in the dashboard to display the data.
    Within this object different categories of items can be used like (list, simple text, charts...)

    Related Pages:

	    https://www.w3schools.com/html/default.asp

    :rtype: :doc:`Components.Components <report/ui>`

    :return: Python HTML object
    """
    if self._ui is None:
      self._ui = Components.Components(self)
    return self._ui

  @property
  def css(self):
    """
    Returns the set of :doc:`CSS Classes <css>` for the HTML report
    """
    return Classes.Catalog(self, {'other': set()})._class_type('other')

  @property
  def js(self):
    """
    Description:
    ------------
    Go to the Javascript section. Property to get all the JavaScript features.
    Most of the standard modules will be available in order to add event and interaction to the Js transpiled

    Usage::

      js.console.log("test")

    Related Pages:

      https://www.w3schools.com/js/default.asp

    :return: Python HTML object
    """
    if self._js is None:
      self._js = js.Js.JsBase(self)
    return self._js

  @property
  def py(self):
    """
    Description:
    ------------
    Python external module section.

    Related Pages:

      https://www.w3schools.com/js/default.asp

    :return: Python HTML object
    """
    if self._py is None:
      self._py = PyExt.PyExt(self)
    return self._py

  @property
  def data(self):
    """
    Description:
    ------------
    Python internal data source management

    This can be extended by inheriting from this epyk.core.data.DataSrc.DataSrc
    and adding extra entry points

    :return: The framework available data source
    """
    return data.Data.DataSrc(self)

  def register(self, ext_components):
    """
    Description:
    ------------

      This function allows you to register external epyk objects (namely coming from Pyk Reports)
      by registering them you this will engrave the object within your report
      The example below will add obj1 and obj2 from an external pyk report previously required,
      then create a div and then add obj3 from an external file

    Usage:
    ------

      rptObj.register([obj1, obj2])
      rptObj.ui.div('this is a div')
      rptObj.register(obj3)
    """
    if type(ext_components) != list:
      ext_components = [ext_components]

    for comp in ext_components:

      if comp.htmlCode in self.components:
        raise Exception("Duplicated Html Code %s in the script !" % comp.htmlCode)

      self.components[comp.htmlCode] = comp

  def get_components(self, htmlcodes):
    """
    Description:
    ------------
    retrieve the components based on their ID.
    This should be used when the htmlCode is defined for a component.

    Attributes:
    ----------
    :param htmlcodes:
    """
    if not isinstance(htmlcodes, list):
      return self.components[htmlcodes]

    return [self.components[htmlcode] for htmlcode in htmlcodes]

  def framework(self, name):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    """
    self._props['context']['framework'] = name.upper()

  @property
  def outs(self):
    """
    Description:
    ------------

    """
    return PyOuts.PyOuts(self)

  @property
  def headers(self):
    """
    Property to the HTML page header
    """
    if self._header_obj is None:
      self._header_obj = html.Header.Header(self)
    return self._header_obj

  def dumps(self, data):
    """
    Description:
    ------------
    Function used to dump the data before being sent to the Javascript layer
    This function relies on json.dumps with a special encoder in order to work with Numpy array and Pandas data structures.

    As NaN is not valid on the Json side those object are not allowed during the dump.
    It is advised to use fillna() in your script before returning the data to the framework to avoid this issue.

    Example::

      report.dumps(result)

    Related Pages:

			https://docs.python.org/2/library/json.html

    Attributes:
    ----------
    :param data: The python dictionary or data structure

    :return: The serialised data
    """
    return json.dumps(data, cls=js.JsEncoder.Encoder, allow_nan=False)



"""
Javascript Interface to the Tabulator Module

http://tabulator.info/docs/4.4/components

TODO: Add tree event on RowComponent
"""

from epyk.core.js import JsUtils
from epyk.core.js.objects import JsNodeDom
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import JsPackage


class Settings(object):
  def __init__(self, src, options):
    self.src = src
    self.__headerVisible = True
    self.__ctx = {}

  @property
  def headerVisible(self):
    """
    By setting the headerVisible option to false you can hide the column headers and present the table as a simple list if needed.
    """
    return self.__headerVisible

  @headerVisible.setter
  def headerVisible(self, flag):
    self.__headerVisible = flag
    self.__ctx['headerVisible'] = flag


class Navigation(JsPackage):
  def prev(self):
    """
    next editable cell on the left, if none available move to the right most editable cell on the row above

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.prev()" % self.toStr())

  def next(self):
    """
    next editable cell on the right, if none available move to left most editable cell on the row below

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.next()" % self.toStr())

  def left(self):
    """
    next editable cell on the left, return false if none available on row

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.left()" % self.toStr())

  def right(self):
    """
    next editable cell on the right, return false if none available on row

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.right()" % self.toStr())

  def up(self):
    """
    move to the same cell in the row above

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.up()" % self.toStr())

  def down(self):
    """
    move to the same cell in the row below

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.down()" % self.toStr())

  def toStr(self):
    """
    Javascript representation

    :return: Return the Javascript String
    """
    if self._selector is None:
      raise Exception("Selector not defined, use this() or new() first")

    if len(self._js) == 0:
      return self._selector

    strData = "%(jqId)s.%(items)s" % {'jqId': self._selector, 'items': ".".join(self._js)}
    self._js = [] # empty the stack
    return JsObjects.JsObject.JsObject.get(strData)


class CellComponent(JsPackage):
  lib_alias = {"js": "tabulator", 'css': "tabulator"}
  lib_selector = "cell"

  def getElement(self):
    """
    The getElement function returns the DOM node for the cell.

    Documentation
    http://tabulator.info/docs/4.4/components

    :rtype: JsNodeDom.JsDoms
    :return:
    """
    return JsNodeDom.JsDoms(varName="%s.getElement()" % self.toStr())

  def getColumn(self):
    """
    The getColumn function returns the ColumnComponent for the column that contains the cell.

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    return ColumnComponent("%s.getColumn()" % self.toStr())

  def getRow(self):
    """
    The getRow function returns the RowComponent for the row that contains the cell.

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    return RowComponent("%s.getRow()" % self.toStr())

  def getData(self):
    """
    The getData function returns the data for the row that contains the cell.

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getData()" % self.toStr())

  def getField(self):
    """
    The getField function returns the field name for the column that contains the cell.

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getField()" % self.toStr())

  def restoreOldValue(self):
    """
    The restoreOldValue reverts the value of the cell back to its previous value, without triggering any of the cell edit callbacks.

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    return JsObjects.JsString.JsString("%s.restoreOldValue()" % self.toStr())

  def getOldValue(self):
    """
    The getOldValue function returns the previous value of the cell. Very usefull in the event of cell update callbacks.

    Documentation
    http://tabulator.info/docs/3.5#component-cell

    :return:
    """
    return JsObjects.JsString.JsString("%s.restoreOldValue()" % self.toStr())

  def getValue(self):
    """
    The getValue function returns the current value for the cell.

    Documentation
    http://tabulator.info/docs/3.5#component-cell

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getValue()" % self.toStr())

  def setValue(self, jsString, jsBoolean=True):
    """
    You can change the value of the cell using the setValue function.
    The first parameter should be the new value for the cell, the second optional parameter will apply the column mutators to the value when set to true (default = true).

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    key = JsUtils.jsConvertData(jsString, None)
    value = JsUtils.jsConvertData(jsBoolean, None)
    return JsObjects.JsObject.JsObject("%s.setValue(%s, %s)" % (self.toStr(), key, value))

  def checkHeight(self):
    """
    If you are making manual adjustments to elements contained withing the cell, or the cell itself, it may sometimes be necessary to recalculate the height of all the cells in the row to make sure they remain aligned

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    return JsObjects.JsBoolean.JsBoolean("%s.checkHeight()" % self.toStr())

  def edit(self, jsBoolean=True):
    """
    You and programatically cause a cell to open its editor element using the edit function.

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    jsBoolean = JsUtils.jsConvertData(jsBoolean, None)
    return JsObjects.JsObject.JsObject("%s.edit(%s)" % (self.toStr(), jsBoolean))

  def cancelEdit(self):
    """
    You and programatically cancel a cell edit that is currently in progress by calling the cancelEdit function.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.cancelEdit()" % self.toStr())

  def nav(self):
    """
    When a cell is being edited it is possible to move the editor focus from the current cell to one if its neighbours
    :return:
    """
    return Navigation("%s.nav()" % self.toStr())


class GroupComponent(JsPackage):
  def getElement(self):
    """
    The getElement function returns the DOM node for the group header.

    Documentation
    http://tabulator.info/docs/3.5#component-cell

    :return:
    """
    return JsNodeDom.JsDoms(varName="%s.getElement()" % self.toStr())

  def getKey(self):
    """
    The getKey function returns the unique key that is shared between all rows in this group.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getKey()" % self.toStr())

  def getField(self):
    """
    The getField function returns the string of the field that all rows in this group have been grouped by. (if a function is used to group the rows rather than a field, this function will return false)

    :return:
    """
    return JsObjects.JsString.JsString("%s.getField()" % self.toStr())

  def getRows(self):
    """
    The getRows function returns an array of RowComponent objects, one for each row in the group.

    :return:
    """
    return RowComponent("%s.getRows()" % self.toStr())

  def getSubGroups(self):
    """
    The getParentGroup function returns the GroupComponent for the parent group of this group. if no parent exists, this function will return false

    :return:
    """
    return JsObjects.JsArray.JsArray("%s.getParentGroup()" % self.toStr())

  def getVisibility(self):
    """
    The getVisibility function returns a boolean to show if the group is visible, a value of true means it is visible.

    :return:
    """
    return JsObjects.JsBoolean.JsBoolean("%s.getVisibility()" % self.toStr())

  def show(self):
    """
    The show function shows the group if it is hidden.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.show()" % self.toStr())

  def hide(self):
    """
    The hide function hides the group if it is visible.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.hide()" % self.toStr())

  def toggle(self):
    """
    The toggle function toggles the visibility of the group, switching between hidden and visible.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.toggle()" % self.toStr())

  def getTable(self):
    """
    The getTable function returns the Tabulator object for the table containing the group.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getTable()" % self.toStr())

  def getParentColumn(self):
    """
    The getParentGroup function returns the GroupComponent for the parent group of this group. if no parent exists, this function will return false

    :return:
    """
    return GroupComponent("%s.getParentGroup()" % self.toStr())


class ColumnComponent(JsPackage):
  lib_alias = {"js": "tabulator", 'css': "tabulator"}
  lib_selector = "column"

  def getElement(self):
    """
    The getElement function returns the DOM node for the column.

    :return:
    """
    return JsNodeDom.JsDoms("%s.getElement()" % self.toStr())

  def getTable(self):
    """
    The getTable function returns the Tabulator object for the table containing the column.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getTable()" % self.toStr())

  def getDefinition(self):
    """
    The getDefinition function returns the column definition object for the column.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getDefinition()" % self.toStr())

  def getField(self):
    """
    The getField function returns the field name for the column.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getField()" % self.toStr())

  def getCells(self):
    """
    The getCells function returns an array of CellComponent objects, one for each cell in the column.

    :return:
    """
    return CellComponent("%s.getCells()" % self.toStr())

  def getNextColumn(self):
    """
    The getNextColumn function returns the Column Component for the next visible column in the table, if there is no next column it will return a value of false.

    :return:
    """
    return ColumnComponent("%s.getNextColumn()" % self.toStr())

  def getPrevColumn(self):
    """
    The getPrevColumn function returns the Column Component for the previous visible column in the table, if there is no previous column it will return a value of false.

    :return:
    """
    return ColumnComponent("%s.getPrevColumn()" % self.toStr())

  def getVisibility(self):
    """
    The getVisibility function returns a boolean to show if the column is visible, a value of true means it is visible.

    :return:
    """
    return JsObjects.JsBoolean.JsBoolean("%s.getVisibility()" % self.toStr())

  def show(self):
    """
    The show function shows the column if it is hidden.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.show()" % self.toStr())

  def hide(self):
    """
    The hide function hides the column if it is visible.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.hide()" % self.toStr())

  def toggle(self):
    """
    The toggle function toggles the visibility of the column, switching between hidden and visible.

    Documentation
    http://tabulator.info/docs/4.5/columns#addColumn

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.hide()" % self.toStr())

  def delete(self):
    """
    The delete function deletes the column, removing it from the table

    :return:
    """
    return self.fnc_closure_in_promise("delete()")

  def scrollTo(self):
    """
    The scrollTo function will scroll the table to the column if it is visible.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.scrollTo()" % self.toStr())

  def move(self, jsString, jsBoolean):
    """
    ou can move a column next to another column using the move function.

    :return:
    """
    jsString = JsUtils.jsConvertData(jsString, None)
    jsBoolean = JsUtils.jsConvertData(jsBoolean, None)
    return JsObjects.JsObject.JsObject("%s.move(%s, %s)" % (self.toStr(), jsString, jsBoolean))

  def getSubColumns(self):
    """
    The getSubColumns function returns an array of ColumnComponent objects, one for each sub column of this column.

    :return:
    """
    return ColumnComponent("%s.getSubColumns()" % self.toStr())

  def getParentColumn(self):
    """
    The getParentColumn function returns the ColumnComponent for the parent column of this column. if no parent exists, this function will return false

    :return:
    """
    return ColumnComponent("%s.getParentColumn()" % self.toStr())

  def headerFilterFocus(self):
    """
    The headerFilterFocus function will place focus on the header filter element for this column if it exists.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.headerFilterFocus()" % self.toStr())

  def setHeaderFilterValue(self, jsString):
    """
    The setHeaderFilterValue function set the value of the columns header filter element to the value provided in the first argument.

    :param jsString:
    :return:
    """
    jsString = JsUtils.jsConvertData(jsString, None)
    return JsObjects.JsObject.JsObject("%s.setHeaderFilterValue(%s)" % (self.toStr(), jsString))

  def reloadHeaderFilter(self):
    """
    The reloadHeaderFilter function rebuilds the header filter element, updating any params passed into the editor used to generate the filter.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.reloadHeaderFilter()" % self.toStr())


class ColumnComponents(JsPackage):
  lib_selector = "column"

  def forEach(self, jsFnc):
    """

    :param jsFnc:
    :return:
    """
    return self.fnc_closure("forEach(function(rec){%s})" % JsUtils.jsConvertFncs(jsFnc, toStr=True))

  @property
  def table(self):
    """
    Return to the parent table
    """
    self._parent._js.append([])
    return self._parent


class RowComponent(JsPackage):
  lib_selector = "row"

  def update(self, data):
    """
    You can update the data in the row using the update function.
    You should pass an object to the function containing any fields you wish to update

    Documentation
    http://www.tabulator.info/docs/4.0/update

    :param data: Dictionary with the data to override

    :return:
    """
    data = JsUtils.jsConvertData(data, None)
    return self.fnc_closure("update(%s)" % data)

  def getData(self):
    """
    The getData function returns the data object for the row.

    Documentation
    http://tabulator.info/docs/4.4/components

    :return:
    """
    return JsObjects.JsArray.JsArray("%s.getData()" % self.toStr())

  def getElement(self):
    """
    The getElement function returns the DOM node for the row.

    Documentation
    http://tabulator.info/docs/4.4/components

    """
    return JsNodeDom.JsDoms(varName="%s.getElement()" % self.toStr())

  def getCells(self):
    """
    The getCells function returns an array of CellComponent objects, one for each cell in the row.

    :return:
    """
    return JsObjects.JsArray.JsArray("%s.getCells()" % self.toStr())

  def getCell(self):
    """
    The getCell function returns the CellComponent for the specified column from this row.

    :return:
    """
    return CellComponent("%s.getCell()" % self.toStr())

  def getIndex(self):
    """
    The getIndex function returns the index value for the row. (this is the value from the defined index column, NOT the row's position in the table)

    :return:
    """
    return JsObjects.JsNumber.JsNumber("%s.getIndex()" % self.toStr())

  def getPosition(self, jsBoolean=True):
    """
    Use the getPosition function to retrieve the numerical position of a row in the table.
    By default this will return the position of the row in all data, including data currently filtered out of the table.

    :return:
    """
    jsBoolean = JsUtils.jsConvertData(jsBoolean, None)
    return JsObjects.JsNumber.JsNumber("%s.getPosition(%s)" % (self.toStr(), jsBoolean))

  def getGroup(self):
    """
    When using grouped rows, you can retrieve the group component for the current row using the getGroup function.

    :return:
    """
    return GroupComponent("%s.getGroup()" % self.toStr())

  def delete(self):
    """
    The delete function deletes the row, removing its data from the table

    Documentation
    http://www.tabulator.info/docs/4.0/update

    :return:
    """
    return self.fnc_closure_in_promise("delete()")

  def scrollTo(self):
    """
    The scrollTo function will scroll the table to the row if it passes the current filters.

    :return:
    """
    return self.fnc_closure("scrollTo()")

  def pageTo(self):
    """
    The pageTo function will load the page for the row if it passes the current filters.

    :return:
    """
    return self.fnc_closure("pageTo()")

  def move(self, jsIndex, jsBoolean=True):
    """
    You can move a row next to another row using the move function.

    :param jsIndex:
    :param jsBoolean:
    :return:
    """
    jsIndex = JsUtils.jsConvertData(jsIndex, None)
    jsBoolean = JsUtils.jsConvertData(jsBoolean, None)
    return self.fnc_closure("move(%s, %s)" % (jsIndex, jsBoolean))

  def select(self):
    """
    The select function will select the current row.

    :return:
    """
    return self.fnc_closure("select()")

  def deselect(self):
    """
    The deselect function will deselect the current row.

    :return:
    """
    return self.fnc_closure("deselect()")

  def toggleSelect(self):
    """
    The toggleSelect function will toggle the selected state the current row.

    :return:
    """
    return self.fnc_closure("toggleSelect()")

  def isSelected(self):
    """
    The isSelected function will return a boolean representing the current selected state of the row.

    :return:
    """
    return JsObjects.JsBoolean.JsBoolean("%s.isSelected()" % self.toStr())

  def normalizeHeight(self):
    """
    If you are making manual adjustments to elements contained within the row, it may sometimes be necessary to recalculate the height of all the cells in the row to make sure they remain aligned

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.normalizeHeight()" % self.toStr())

  def reformat(self):
    """
    If you want to re-format a row once it has been rendered to re-trigger the cell formatters and the rowFormatter callback, Call the reformat function.

    :return:
    """
    return self.fnc_closure("reformat()")

  def freeze(self):
    """
    You can freeze a row at the top of the table by calling the freeze function. This will insert the row above the scrolling portion of the table in the table header.

    :return:
    """
    return self.fnc_closure("freeze()")

  def unfreeze(self):
    """
    A frozen row can be unfrozen using the unfreeze function. This will remove the row from the table header and re-insert it back in the table.

    :return:
    """
    return self.fnc_closure("unfreeze()")


class Tabulator(JsPackage):
  lib_alias = {"js": "tabulator", 'css': "tabulator"}

  def previousPage(self):
    """
    Description:
    -----------

    http://tabulator.info/docs/4.1/page
    """
    return JsObjects.JsPromise("%s.previousPage()" % self.varId)

  def nextPage(self):
    """
    Description:
    -----------

    http://tabulator.info/docs/4.1/page

    :return:
    """
    return JsObjects.JsPromise("%s.nextPage()" % self.varId)

  def setPage(self, i):
    """
    Description:
    -----------

    http://tabulator.info/docs/4.1/page

    :param i:
    :return:
    """
    return JsObjects.JsPromise("%s.setPage(%s)" % (self.varId, i))

  def setPageSize(self, i):
    """
    Description:
    -----------

    http://tabulator.info/docs/4.1/page

    Attributes:
    ----------
    :param i:
    """
    return JsObjects.JsPromise("%s.setPageSize(%s)" % (self.varId, i))

  def setGroupBy(self, column):
    """
    Description:
    -----------
    You can use the setGroupBy function to change the fields that rows are grouped by.
    This function has one argument and takes the same values as passed to the groupBy setup option.

    Related Pages:

			http://www.tabulator.info/docs/4.1/group

    Attributes:
    ----------
    :param column:
    """
    return JsObjects.JsVoid("%s.setGroupBy(%s)" % (self.varId, JsUtils.jsConvertData(column, None)))

  def setGroupStartOpen(self, flag):
    """
    You can use the setGroupStartOpen function to change the default open state of groups.
    This function has one argument and takes the same values as passed to the groupStartOpen setup option.

    Documentation
    http://www.tabulator.info/docs/4.1/group

    :param flag:

    :return:
    """
    return JsObjects.JsVoid("%s.setGroupStartOpen(%s)" % (self.varId, JsUtils.jsConvertData(flag, None)))

  def setGroupHeader(self, jsFnc):
    """
    You can use the setGroupHeader function to change the header generation function for each group.
    This function has one argument and takes the same values as passed to the groupHeader setup option.

    Documentation
    http://www.tabulator.info/docs/4.1/group

    :param jsFnc:

    :return:
    """
    return self.fnc_closure("setGroupHeader(function(value, count, data, group){%s})" % JsUtils.jsConvertFncs(jsFnc, toStr=True))

  def deleteRow(self, n):
    """
    You can delete any row in the table using the deleteRow function.
    The first argument is the row you want to delete, it will take any of the standard row component look up options.

    documentation
    http://www.tabulator.info/docs/4.0/update

    :param n:

    :return:
    """
    return JsObjects.JsPromise("%s.deleteRow(%s)" % (self.varId, n))

  def addRow(self, data, flag=False):
    """
    You can add a row to the table using the addRow function.

    The first argument should be a row data object. If you do not pass data for a column, it will be left empty.
    To create a blank row (ie for a user to fill in), pass an empty object to the function.

    The second argument is optional and determines whether the row is added to the top or bottom of the table.
    A value of true will add the row to the top of the table, a value of false will add the row to the bottom of the table.
    If the parameter is not set the row will be placed according to the addRowPos global option.

    Documentation
    http://tabulator.info/docs/4.3/update#addrow

    :param data:
    :param flag:
    """
    data = JsUtils.jsConvertData(data, None)
    flag = JsUtils.jsConvertData(flag, None)
    return self.fnc_closure("addRow(%s, %s)" % (data, flag))

  def updateRow(self, rowId, data):
    """
    The updateRow and row.updatemethods return a promise,
    this can be used to run any other commands that have to be run after the data has been loaded into the table.

    Documentation
    http://www.tabulator.info/docs/4.0/update

    :param rowId:
    :param data:

    :return:
    """
    return self.fnc_closure("updateRow(%s, %s)" % (rowId, JsUtils.jsConvertData(data, None)))

  def updateOrAddRow(self, rowId, data):
    """
    If you don't know whether a row already exists you can use the updateOrAddRow function.
    This will check if a row with a matching index exists, if it does it will update it, if not it will add a new row with that data

    Documentation
    http://www.tabulator.info/docs/4.0/update

    :param rowId:
    :param data:

    :return:
    """
    return self.fnc_closure("updateOrAddRow(%s, %s)" % (rowId, JsUtils.jsConvertData(data, None)))

  def getRow(self, jsIndex):
    """
    Get the Row component

    Documentation
    http://tabulator.info/docs/4.1/components

    Example
    var row = cell.getRow();
    """
    row = RowComponent(self.src, selector="getRow(%s)" % jsIndex, setVar=False, parent=self)
    self.fnc(row)
    return row

  def getSelectedRows(self):
    """
    To get the RowComponent's for the selected rows at any time you can use the getSelectedRows function.

    Documentation
    http://tabulator.info/docs/4.0/select

    :return:
    """
    return JsObjects.JsArray.JsArray("%s.getSelectedRows()" % self.varId)

  def getRows(self):
    """
    The getRows function returns an array of RowComponent objects, one for each row in the table.

    Example
    rows = table.getRows()

    Documentation
    http://tabulator.info/docs/4.1/components

    :return:
    """
    return JsObjects.JsArray.JsArray("%s.getRows()" % self.varId)

  def getRowPosition(self, row, bool=True):
    """
    The new getRowPosition function and Row Component getPosition function allow you to retrieve the current position of a row in the table.

    Documentation
    http://tabulator.info/news

    :return:
    """
    return JsObjects.JsNumber.JsNumber("%s.getRowPosition(%s, %s)" % (self.varId, row, bool))

  def getPageSize(self):
    """

    http://tabulator.info/docs/4.1/page

    :return:
    """
    return JsObjects.JsNumber.JsNumber("%s.getPageSize()" % self.varId)

  def getPage(self):
    """

    http://tabulator.info/docs/4.1/page

    :return:
    """
    return JsObjects.JsNumber.JsNumber("%s.getPage()" % self.varId)

  def getPageMax(self):
    """
    http://tabulator.info/docs/4.1/page

    :return:
    """
    return JsObjects.JsNumber.JsNumber("%s.getPageMax()" % self.varId)

  def getRowFromPosition(self, n, bool=True):
    """

    :param n:
    """
    bool = JsUtils.jsConvertData(bool, None)
    return JsObjects.JsNumber.JsNumber("%s.getRowFromPosition(%s, %s)" % (self.varId, n, bool))

  def toggleColumn(self, column):
    """
    You can toggle the visibility of a column at any point using the toggleColumn function. Pass the field name of the column you wish to toggle as the first parameter of the function.

    Example
    table.toggleColumn("name")

    Documentation
    http://tabulator.info/docs/4.5/columns#addColumn

    :return:
    """
    column = JsUtils.jsConvertData(column, None)
    return JsObjects.JsPromise("%s.toggleColumn(%s)" % (self.varId, column))

  def hideColumn(self, column):
    """
    You can hide a visible column at any point using the hideColumn function. Pass the field name of the column you wish to hide as the first parameter of the function.

    Example
    table.hideColumn("name")

    Documentation
    http://tabulator.info/docs/4.5/columns#addColumn

    :param column:
    """
    column = JsUtils.jsConvertData(column, None)
    return JsObjects.JsPromise("%s.hideColumn(%s)" % (self.varId, column))

  def hideColumns(self, columna):
    """
    You can hide a visible column at any point using the hideColumn function. Pass the field name of the column you wish to hide as the first parameter of the function.

    Example
    table.hideColumn("name")

    Documentation
    http://tabulator.info/docs/4.5/columns#addColumn

    :param column:
    """
    columna = JsUtils.jsConvertData(columna, None)
    return JsObjects.JsPromise("%s.forEach(function(c){%s.hideColumn(c)})" % (columna, self.varId))

  @property
  def getColumns(self):
    """
    To get an array of Column Components for the current table setup, call the getColumns function.
    This will only return actual data columns not column groups.

    Example
    var cols = table.getColumns()

    Documentation
    http://tabulator.info/docs/4.5/columns#getColumns

    """
    columns = ColumnComponents(self.src, selector="getColumns()", setVar=False, parent=self)
    self.fnc(columns)
    return columns

  def addColumn(self, jsData, before=False, position=""):
    """
    If you wish to add a single column to the table, you can do this using the addColumn function

    Example
    table.addColumn({"title": "Age", "field": "age"}, True, "name");

    Documentation
    http://tabulator.info/docs/4.5/columns#addColumn

    :param jsData: The column definition object for the column you want to add
    :param before: Determines how to position the new column.
                   A value of true will insert the column to the left of existing columns, a value of false will insert it to the right
    :param position: The field to insert the new column next to, this can be any of the standard column component look up options.
    """
    jsData = JsUtils.jsConvertData(jsData, None)
    before = JsUtils.jsConvertData(before, None)
    position = JsUtils.jsConvertData(position, None)
    return JsObjects.JsPromise("%s.addColumn(%s, %s, %s)" % (self.varId, jsData, position, before))

  def deleteColumn(self, jsData):
    """
    To permanently remove a column from the table deleteColumn function. This function takes any of the standard column component look up options as its first parameter.

    Documentation
    http://tabulator.info/docs/4.5/columns#delete

    :param jsData:
    """
    return JsObjects.JsPromise("%s.deleteColumn(%s)" % (self.varId, JsUtils.jsConvertData(jsData, None)))

  def redraw(self, jsData):
    """
    This can be done by calling the redraw method. For example, to trigger a redraw whenever the viewport width is changed:

    The redraw function also has an optional boolean argument that when set to true triggers a full rerender of the table including all data on all rows.

    Documentation
    http://tabulator.info/docs/4.5/layout#redraw

    :param jsData: Boolean. Trigger full rerender including all data and rows
    :return:
    """
    return JsObjects.JsVoid("%s.redraw(%s)" % (self.varId, JsUtils.jsConvertData(jsData, None)))

  def blockRedraw(self):
    """
    To get around this you can use the blockRedraw and restoreRedraw functions to temporarlity disable all table redraw actions while you are manipulating the table data.

    Example
    table.blockRedraw(); //block table redrawing

    Documentation
    http://tabulator.info/docs/4.5/release#redraw-block
    """
    return JsObjects.JsVoid("%s.blockRedraw()" % self.varId)

  def restoreRedraw(self):
    """

    Example
    table.restoreRedraw(); //restore table redrawing

    Documentation
    http://tabulator.info/docs/4.5/release#redraw-block
    """
    return JsObjects.JsVoid("%s.restoreRedraw()" % self.varId)

  def setSort(self, jsData):
    """

    Example
    table.setSort([{column:"age", dir:"asc"}]);

    Documentation
    http://tabulator.info/docs/4.0/sort

    :param jsData:
    """
    return JsObjects.JsVoid("%s.setSort(%s)" % (self.varId, JsUtils.jsConvertData(jsData, None)))

  def getGroups(self, jsData=None):
    """
    You can use the getGroups function to retrieve an array of all the first level Group Components in the table.

    Example
    var cols = table.getColumns()

    Documentation
    http://tabulator.info/docs/4.0/group

    :param jsData: To get a structured array of Column Components that includes column groups, pass a value of true as an argument
    :return:
    """
    if jsData is not None:
      return JsObjects.JsObject.JsObject("%s.getGroups(%s)" % (self.varId, JsUtils.jsConvertData(jsData, None)))

    return JsObjects.JsObject.JsObject("%s.getGroups()" % self.varId)

  def clearData(self):
    """
    You can remove all data from the table using the clearData function:

    Example
    table.clearData()

    Documentation
    http://tabulator.info/docs/4.0/update

    :return:
    """
    return JsObjects.JsVoid("%s.clearData()" % self.varId)

  def showColumn(self, column):
    """
    Description:
    ------------
    You can show a hidden column at any point using the showColumn function. Pass the field name of the column you wish to show as the first parameter of the function.

    Related Pages:

			http://tabulator.info/docs/4.0/columns

    Attributes:
    ----------
    :param column: String. The column name to be displayed
    """
    return JsObjects.JsPromise("%s.showColumn(%s)" % (self.varId, JsUtils.jsConvertData(column, None)))

  def showColumns(self, columna):
    """
    Description:
    ------------
    You can show a hidden columns at any point using the showColumn function. Pass the field name of the column you wish to show as the first parameter of the function.

    Example
    table.showColumna("name")

    Related Pages:

			http://tabulator.info/docs/4.0/columns

    Attributes:
    ----------
    :param columna: List. The column names to be displayed
    """
    return JsObjects.JsPromise("%s.forEach(function(c){%s.showColumn(c)})" % (JsUtils.jsConvertData(columna, None), self.varId))

  def setData(self, data):
    """

    :return:
    """
    return JsObjects.JsVoid("%s.setData(%s)" % (self.varId, JsUtils.jsConvertData(data, None)))

  def replaceData(self, jsData=None):
    """
    The replaceData function lets you silently replace all data in the table without updating scroll position, sort or filtering, and without triggering the ajax loading popup

    table.replaceData() //trigger reload of ajax data from ajaxURL property

    :return:
    """
    if jsData is None:
      return JsObjects.JsObject.JsObject("%s.replaceData()" % self.varId)

    return JsObjects.JsObject.JsObject("%s.replaceData(%s)" % (self.varId, JsUtils.jsConvertData(jsData, None)))

  def getData(self):
    """
    You can retrieve the data stored in the table using the getData function.

    :return:
    """
    return JsObjects.JsObject.JsObject("%s.getData()" % self.varId)
